/*
 FCEU.h

 * MacFCEU 0.6.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _FCEUH

/* ---- constants */

#define FCEU_VERSION_NUMERIC	9815

#define DECLFR(x)		uint8 FP_FASTAPASS(1) x (uint32 A)
#define DECLFW(x)		void FP_FASTAPASS(2) x (uint32 A, uint8 V)

 /* twelfth-root-two */
#define M_12R2			(double)1.059463094359295264562

#define GI_RESETM2		1
#define GI_POWER		2
#define GI_CLOSE		3

#define JOY_A			1
#define JOY_B			2
#define JOY_SELECT		4
#define JOY_START		8
#define JOY_UP			0x10
#define JOY_DOWN		0x20
#define JOY_LEFT		0x40
#define JOY_RIGHT		0x80

/* ---- typedefs */

typedef struct FCEUS
{
	int	PAL;
	int	NetworkPlay;
	int	SoundVolume;
	int	GameGenie;

	/* Current first and last rendered scanlines. */
	int	FirstSLine, LastSLine;

	/* Driver code(user)-specified first and last rendered scanlines. 
	 Usr*SLine[0] is for NTSC, Usr*SLine[1] is for PAL.
	*/
	int	UsrFirstSLine[2];
	int	UsrLastSLine[2];
	int	UsrVBLSLines[2]; /* added; 2010.01.04 */
	int	SnapName;
	uint32	SndRate;
	int	soundq;
	int	lowpass;
	int	soundrectamp; /* added; 2010.03.23 */
	uint8	soundactivech; /* moved; 2010.03.23 */
} FCEUS;

/* ---- globals */

extern struct FCEUGI *FCEUGameInfo;

extern FCEUS FSettings;

extern int fceuindbg;

extern uint8 ConsolePitchFix;
extern uint32 EmulationSpeedRatio;

extern uint64 timestampbase;

extern uint8 RAM[0x800];
extern uint8 GameMemBlock[131072];

extern readfunc ARead[0x10000];
extern writefunc BWrite[0x10000];

extern void (*GameInterface)(int h);
extern void (*GameStateRestore)(int version);

extern uint8 pale;
extern uint8 vsdip;

extern int GameAttributes;
extern uint8 PAL;

/* ---- prototypes */

/* made public for checking purposes; 2010.04.01 04:55 GMT */
DECLFW(BNull);
DECLFR(ANull);

void ResetGameLoaded(void);

void FCEU_MemoryRand(uint8 *ptr,
	uint32 size);
writefunc FASTAPASS(1) GetWriteHandler(int32 a);
readfunc FASTAPASS(1) GetReadHandler(int32 a);

void FASTAPASS(3) SetReadHandler(int32 start,
	int32 end,
	readfunc f);
void FASTAPASS(3) SetWriteHandler(int32 start,
	int32 end,
	writefunc f);
/*you have to pass the correct final address for the custom-incremental ones;
 they set from end
*/
void FASTAPASS(4) SetReadHandlerI(int d,
	int32 start,
	int32 end,
	readfunc f);
void FASTAPASS(4) SetWriteHandlerI(int d,
	int32 start,
	int32 end,
	writefunc f);

void FCEU_ResetVidSys(void);

void ResetMapping(void);
void ResetNES(void);
void PowerNES(void);

/* these previously had 'format' as char*; 2010.01.05 */
void FCEU_PrintError(const char *format, ...);
void FCEU_printf(const char *format, ...);
void FCEU_DispMessage(const char *format, ...);

void SetNESDeemph(uint8 d, int force);

void DrawTextTrans(uint8 *dstpix,
	uint32 width,
	const uint8 *textmsg,
	uint8 fgcolor);

void FCEU_PutImage(void);

#ifdef FRAMESKIP
//void FCEU_PutImageDummy(void);
#endif

 #define _FCEUH
#endif
