/*
 PPU.c
  Picture Processing Unit 2C02/2C07

 * MacFCEU 0.6.1
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  ppu.c Copyright (C) 1998 BERO
 *  ppu.c Copyright (C) 2003 Xodnizel
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or   
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "types.h"
#include "x6502.h"
#include "FCEU.h"
#include "git.h"
#include "PPU.h"
#include "NSF.h"
#include "GameGenie.h"
#include "fsound.h"
#include "general.h"
#include "memory.h"

#include "cart.h"
#include "palette.h"
#include "state.h"
#include "video.h"
#include "input.h"
#include "boards/MMC5.h"

#define VBlankON		(PPU[0]&0x80)   /* Generate VBlank NMI */
#define Sprite16		(PPU[0]&0x20)   /* Sprites 8x16/8x8 */
#define BGAdrHI			(PPU[0]&0x10)   /* BG pattern adr $0000/$1000 */
#define SpAdrHI			(PPU[0]&0x08)   /* Sprite pattern adr $0000/$1000 */
#define IncByV			(PPU[0]&0x04)   /* by 32, otherwise horiz. (1) */

#define Deempbits		0xE0		/* deemphasis bits */
#define Deemphas		(PPU[1]&0xE0)
#define SpriteON		(PPU[1]&0x10)   /* show sprite layer */
#define ScreenON		(PPU[1]&0x08)   /* show screen */
#define RenderON		(PPU[1]&0x18)	/* either */
#define SpFirst8		(PPU[1]&0x04)	/* render first 8 pixels on line */
#define BGFirst8		(PPU[1]&0x02)	/* render first 8 pixels on line */
#define GrayLine		(PPU[1]&0x01)	/* erase line data (leave mask) */

#define PPU_status		(PPU[2])

#ifdef C80x86
 #define tilespr32(P,R,d)	do{uint8 *S=R; uint32 V=d; \
	 asm volatile( \
	 "movl %%ebx,%%eax\n\t" \
	 "andl $15,%%eax\n\t" \
	 "movb (%%esi,%%eax),%%cl\n\t" \
	 "shrl $4, %%ebx\n\t" \
	 "movb %%cl, (%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 1(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 2(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 3(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 4(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 5(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 6(%%edi)\n\t" \
\
         /*"movl %%ebx,%%eax\n\t"*/ \
         /*"andl $15,%%eax\n\t"*/ \
         /*"movb (%%esi,%%eax),%%cl\n\t"*/ \
         /*"movb %%cl, 7(%%edi)\n\t"*/ \
	 "movb (%%esi, %%ebx),%%cl\n\t" \
	 "movb %%cl, 7(%%edi)\n\t" \
	 : \
	 : "S" (S), "D" (P), "b" (V) \
	 : "eax", "cl" ); \
}while(0)
#elif defined(env68k)
/* P cannot be advanced from here so preincrementing addressing mode won't fly */
 #pragma parameter tilespr32(__A0,__A1,__D0)
 extern void tilespr32(uint8 *P, uint8 *R, uint32 d) =
 {
	0x720F,				//	MOVEQ		#15,D1
	0xC280,				//	AND.L		D0,D1
	0x3431,0x1800,			//	MOVE.W		0(A1,D1.L),D2
	0xE9C0,0x1604,			//	BFEXTU		D0{24:4},D1
	0x1431,0x1800,			//	MOVE.B		0(A1,D1.L),D2
	0x4842,				//	SWAP		D2
	0xE9C0,0x1504,			//	BFEXTU		D0{20:4},D1
	0x3431,0x1800,			//	MOVE.W		0(A1,D1.L),D2
	0xE9C0,0x1404,			//	BFEXTU		D0{16:4},D1
	0x1431,0x1800,			//	MOVE.B		0(A1,D1.L),D2
	0x20C2,				//	MOVE.L		D2,(A0)+
	0xE9C0,0x1304,			//	BFEXTU		D0{12:4},D1
	0x3431,0x1800,			//	MOVE.W		0(A1,D1.L),D2
	0xE9C0,0x1204,			//	BFEXTU		D0{8:4},D1
	0x1431,0x1800,			//	MOVE.B		0(A1,D1.L),D2
	0x4842,				//	SWAP		D2
	0xE9C0,0x1104,			//	BFEXTU		D0{4:4},D1
	0x3431,0x1800,			//	MOVE.W		0(A1,D1.L),D2
	0xE9C0,0x1004,			//	BFEXTU		D0{0:4},D1
	0x1431,0x1800,			//	MOVE.B		0(A1,D1.L),D2
	0x2082				//	MOVE.L		D2,(A0)
 };
#else
 #define tilespr32(P,R,d)	do{ \
	uint8 *S=R; uint32 V=d; \
	*P=S[V&0xF]; \
	P[1]=S[(V>>=4)&0xF]; \
	P[2]=S[(V>>=4)&0xF]; \
	P[3]=S[(V>>=4)&0xF]; \
	P[4]=S[(V>>=4)&0xF]; \
	P[5]=S[(V>>=4)&0xF]; \
	P[6]=S[(V>>=4)&0xF]; \
	P[7]=S[V>>=4]; \
}while(0)
#endif

static void FASTAPASS(1) RefreshLine(int lastpixel);

uint32		PPU_LUT1to4[256]; /* 1024B */
uint32		PPU_TileLUT[128]; /* 512B */
uint8		bitreverseLUT[256];

static void makeluts(uint32 *l1b, uint32 *tlut, uint8 *brt)
{
	{
		uint32	*d = l1b, x = 0, v;

		do
		{ /* unrolled */
			v = x>>7;
			v|= x>>6-4&0x10;
			v|= x<<8-5&0x100;
			v|= x<<12-4&0x1000;
			v|= x<<16-3&0x10000;
			v|= x<<20-2&0x100000;
			v|= x<<24-1&0x1000000;
			*d++ = v,v|= 0x10000000, *d++ = v;
		} while((x+=2)<256);
	}

	{
		uint32	*d = tlut, cc = 0, xo;

		do
		{
			xo = 0;
			do
			{
				uint32	i = 8, k = xo, v = 0, a = 2;

				do
					v|= ((cc>>(k>>3)*2)&3)<<a;
				while(++k,a+=4,--i!=0);
				*d++=v;
			} while(++xo<8);
		} while(++cc<16);
	}

	{
		uint8	*d = brt, v;
		uint32	x = 0;

		do
		{
			v = x<<5&0x40;
			v|= x<<3&0x20;
			v|= (x+x)&0x10;
			v|= x>>1&8;
			v|= x>>3&4;
			v|= x>>5&2;
			v|= x>>7;
			*d++ = v, v|= 0x80;
			*d++ = v;
		} while((x+=2)<256);
	}
}

int		fceuindbg = 0;

uint8		PPUmaxsprites = 8;
uint8		PPUrendis = 0;

static uint8	PPUBoot; /*int ppudead; */
static int	DummyLineEndCC = 15; //kook

uint8		VRAMBuffer, PPUGenLatch;
uint8		*vnapage[4];
uint8		PPUNTARAM;
uint8		PPUCHRRAM;

/* Color [de]emphasis emulation.  Joy... */
uint8		PPUemp = 0;

void (*GameHBIRQHook)(void), (*GameHBIRQHook2)(void);
void FP_FASTAPASS(1) (*PPU_hook)(uint32 A);

/* The makeup of a 15 bit daisy chain:
 bits 12 : fv.3	- fine vertical scroll
 bit  11 : v.1	- vertical scroll counter
 bit  10 : h.1	- horizontal scroll counter
 bits  5 : vt.5	- vertical tile position   -- bit #1 of this shifts palette by 4
 bits  0 : ht.5	- horizontal tile position -- bit #1 of this shifts palette by 2
 bits(-3): fh.3	- fine horizontal scroll
*/
#define HV_SCR_SHIFT		10
#define HT_POS_MASK		0x1F
#define H_POS_MASKS		0x41F
#define T_POS_MASKS		0x3FF
#define POS_MASKS		0x3FFF

uint32	TempAddr, RefreshAddr;

static uint16	ppu_slatch = 0; /* playfield pattern selection latch (1 bit used) */

uint8	vtoggle = 0;
uint8	XOffset = 0;

/* scanline is equal to the current visible scanline we're on. */

int	scanline;
uint32	ppuvbl_scanline_cc; //has FIRSTCYC already subtracted
#define FIRSTCYC 20 //was 12

uint8	PPU[4];
uint8	PPUSPL;

//OAM rewrite pending; but for now, this FCEUX-based array
uint8	UPALRAM[16] = {0xFF}; //4 bytes used; [0] is bg void color

uint8	NTARAM[0x800], PALRAM[0x20];
uint8	SPRAM[0x120], SPRBUF[0x120]; //gearing for accurate emulation; 2010.01.16


#define MMC5SPRVRAMADR(V)	&MMC5SPRVPage[(V)>>HV_SCR_SHIFT][(V)]
#define MMC5BGVRAMADR(V)	((Sprite16||MMC5ABMode!=0) ? \
	MMC5SPRVRAMADR(V):&MMC5BGVPage[(V)>>HV_SCR_SHIFT][(V)])
#define VRAMADR(V)		&VPage[(V)>>HV_SCR_SHIFT][(V)]

#ifdef env68k
DECLFR(PPUA2002);
DECLFR(PPUA200x);
DECLFR(PPUA2007);

DECLFW(PPUB2000);
DECLFW(PPUB2001);
DECLFW(PPUB2002);
DECLFW(PPUB2003);
DECLFW(PPUB2004);
DECLFW(PPUB2005);
DECLFW(PPUB2006);
DECLFW(PPUB2007);
DECLFW(PPUB4014);
 #define STATIC
#else
 #define STATIC static
#endif

STATIC DECLFR(PPUA2002)
{
	uint8 ret;
(void)A;
	FCEUPPU_LineUpdate();
	ret = PPU_status|(PPUGenLatch&0x1F);

	if (fceuindbg)
		return ret;

	vtoggle = 0;
	PPU_status&= 0x7F;
	PPUGenLatch = ret;

	return ret;
}

STATIC DECLFR(PPUA200x) /* Not correct for $2004 reads. */
{
(void)A;
	FCEUPPU_LineUpdate();
	return PPUGenLatch;
}

/*
static DECLFR(A2004)
{
 uint8 ret;

 FCEUPPU_LineUpdate();
 ret = SPRAM[PPU[3]];

 if(PPUSPL>=8)
 {
  if (PPU[3]>=8)
   ret = SPRAM[PPU[3]];
 } else
 {
  //printf("$%02x:$%02x\n",PPUSPL,V);
  ret = SPRAM[PPUSPL];
 }
 PPU[3]++;
 PPUSPL++;
 PPUGenLatch = ret;
 printf("%d, %02x\n",scanline,ret);
 return(ret);
}
*/

#define VRead(d,a) do{ \
	if ((a)<0x2000) \
		d=VPage[(a)>>HV_SCR_SHIFT][a]; \
	else \
		d=vnapage[(a)>>HV_SCR_SHIFT&3][(a)&0x3FF]; \
}while(0)

STATIC DECLFR(PPUA2007)
{
	uint32	raddr = RefreshAddr&POS_MASKS;
	uint8	ret;
(void)A;
	FCEUPPU_LineUpdate();

	ret = VRAMBuffer;

	if (fceuindbg)
		return ret;

	if (PPU_hook)
		PPU_hook(raddr);
	PPUGenLatch = VRAMBuffer;

	//added (again); 2010.03.08
	if ((~raddr&0x3F00)==0)
	{
		if (raddr&3)
			ret = PALRAM[raddr&0x1F];
		else if ((raddr&0xC)==0) //was &=0xC (stupid me); 2010.03.28
			ret = PALRAM[0];
		else
			ret = UPALRAM[raddr&0xC];

		if (GrayLine)
			ret&= 0x30;
		raddr-= 0x1000;
	}

	VRead(VRAMBuffer,raddr);

	if (IncByV)
		RefreshAddr+= 32;
	else
		++RefreshAddr;
	if (PPU_hook)
		PPU_hook(RefreshAddr&0x3fff);

	return ret;
}

STATIC DECLFW(PPUB2000) //hv latch
{
	int		sv;
(void)A;
	FCEUPPU_LineUpdate();
	PPUGenLatch = V;
	if (!VBlankON && (V&PPU_status&0x80))
		XFrame.IRQlow|= FCEU_IQNMI2; //TriggerNMI2
	/*else if (scanline==241 && (V&0x80)==0 && ppu_cycle<3)
		XFrame.IRQlow&=~FCEU_IQNMI2;*/

	PPU[0] = V;
	sv = V<<HV_SCR_SHIFT;
	TempAddr&= 0xF3FF;
	TempAddr|= sv&0xC00;

	ppu_slatch = sv&0x4000;
}

STATIC DECLFW(PPUB2001) //general latch
{
(void)A;
	FCEUPPU_LineUpdate();
	PPUGenLatch = PPU[1] = V;
	if (V&=Deempbits)
		PPUemp = V>>=5; //gets set to PPU[1]>>5 later anyway
}

STATIC DECLFW(PPUB2002) //status
{
(void)A;
	PPUGenLatch = V;
}

STATIC DECLFW(PPUB2003) //OAM address
{
(void)A;
	PPUGenLatch = V;
	PPU[3] = V;
	PPUSPL = V&7;
}

STATIC DECLFW(PPUB2004) //OAM data
{
(void)A;
	PPUGenLatch = V;
	if (PPUSPL<8)
	{
		//printf("$%02x:$%02x\n",PPUSPL,V);
		SPRAM[PPUSPL] = V;
	} else if (PPU[3]>=8)
		SPRAM[PPU[3]] = V;
	++PPU[3];
	++PPUSPL;
}

STATIC DECLFW(PPUB2005) //scroll-toggle h/v
{
	uint32	tmpa = TempAddr;
(void)A;
	FCEUPPU_LineUpdate();
	PPUGenLatch = V;
	if (vtoggle^=1)
		tmpa&= 0xFFE0, tmpa|= V>>3, XOffset = V&7;
	else
		tmpa&= 0x8C1F, tmpa|= (V&~7)<<2, tmpa|= (V&7)<<12;
	TempAddr = tmpa;
}

STATIC DECLFW(PPUB2006) //vertical address
{
(void)A;
	FCEUPPU_LineUpdate();
	PPUGenLatch = V;
	if (vtoggle^=1)
		TempAddr&= 0x00FF, TempAddr|= (V&0x3F)<<8;
	else
	{
		TempAddr&= 0xFF00, RefreshAddr = (TempAddr|=V);
		if (PPU_hook)
			PPU_hook(RefreshAddr);
	}
}

STATIC DECLFW(PPUB2007) //vertical data
{
	uint32		k, addrcode = RefreshAddr&POS_MASKS;
(void)A;
	PPUGenLatch = V;
	if (addrcode>=0x3F00)
	{
		// hmmm....
		V&= 0x3F;
		if ((addrcode&15)==0)
			PALRAM[0x0] = PALRAM[0x4] = PALRAM[0x8] = PALRAM[0xC] = V;
		else if (addrcode&3)
			PALRAM[addrcode&0x1F] = V;
		else
			UPALRAM[addrcode&0xC] = V;
	} else if (addrcode<0x2000)
	{
		k = addrcode>>HV_SCR_SHIFT;
		if (PPUCHRRAM&(1<<k))
			VPage[k][addrcode] = V;
	} else
	{
		k = (addrcode>>HV_SCR_SHIFT)&3;
		if (PPUNTARAM&(1<<k))
			vnapage[k][addrcode&0x3FF] = V;
	}

	if (IncByV)
		addrcode+= 32;
	else
		++addrcode;
	RefreshAddr = addrcode;
	if (PPU_hook)
		PPU_hook(addrcode&POS_MASKS);
}

STATIC DECLFW(PPUB4014)
{
	uint32	t = V<<8;
	int	x = 256;
(void)A;
	do
		X6502_DMW(0x2004,X6502_DMR(t));
	while(++t,--x!=0);
}

#define GETLASTPIXEL		(PAL?(timestamp*48-linestartts)/15:(timestamp*48-linestartts)>>4)

uint8		*PPU_Pline, *PPU_Plinef;
static int	firsttile;
static int	linestartts;
static uint8	endofline = 0;

void FCEUPPU_LineUpdate(void)
{
	if (fceuindbg)
		return;

	if (PPU_Pline)
	{
		int	l = GETLASTPIXEL;

		RefreshLine(l);
	}
}

/* (incomplete) OAM emulation */

uint8		PPUsprlinebuf[256+8];
int32		PPUsphitx;
uint8		PPUsphitdata;

void CheckSpriteHit(int e);
void CheckSpriteHit(int e)
{
	const uint8	*r;
	int		x = PPUsphitx;
	uint8		d;

	/* optimized version using Dan Weiss's suggestion; 2010.04.08 */
	if ((e-=16)>254)
		e = 255; /* end before 255; should fix "Battletoads" */
	if ((x+=8)<e) /* take and pin, 8 maximum */
		e = x;
	if ((x-=8)>=e)
		return;

	r = PPU_Plinef;
	if ((d=PPUsphitdata)&0x80 && !(r[x]&0x40))
		goto hit;
	if (++x>=e)
		return;
	do
	{
		if (!((d+=d)&0x80) || r[x]&0x40)
			continue;
	hit:
		PPU_status|= 0x40;
		PPUsphitx = 0x100;
		break;
	} while(++x<e);
}

uint8		PPUspork = 0; /* spork the world. */
/*	Any sprites on this line?
	Then this will be set to 1.  Needed for zapper
	emulation and *gasp* sprite emulation.
*/

#define V_FLIP	0x80
#define H_FLIP	0x40
#define SP_BACK	0x20

typedef struct SPR
{
        uint8	y, no, atr, x;
} SPR;

typedef struct SPRB
{
        uint8	ca[2], atr, x;
} SPRB;


uint8	PPUnumsprites, SpriteBlurp;

static void FetchSpriteData(void)
{
	SPRB	dst;
	SPR	*spr = (SPR*)SPRAM;
	uint32	Pf = Sprite16>>1;
	int	vofs = (unsigned)((SpAdrHI*2)&~Pf)<<8;
	uint32	H = 8+(Pf>>1);
	int	n = 64;
	int	ns = 0, ms = PPUmaxsprites;
	uint8	sb = 0;

	if (PPU_hook==0)
	{
		do
		{
			unsigned	v = scanline-spr->y;

			if (v>=H)
				continue;
			//printf("%d, %u\n",scanline,(unsigned int)(scanline-spr->y));
			if (ns>=ms)
				{PPU_status|= 0x20; break;}
			if (n==64) /*n==63 [n was 0-63 in `for`]*/
				sb = 1;

			{
				uint8		*C;
				unsigned int	vadr = spr->no;

				if (Sprite16)
					vadr = ((vadr&1)<<12) + ((vadr&0xFE)<<4);
				else
					vadr = (vadr<<4)+vofs;

				if (spr->atr&V_FLIP)
					vadr+= 7, vadr-= v, vadr+= Pf, vadr-= v&8;
				else
					vadr+= v, vadr+= v&8;

				/* Fix this geniestage hack */
				if (MMC5Ex && GenieStage!=1)
					C = MMC5SPRVRAMADR(vadr);
				else
					C = VRAMADR(vadr);

				dst.ca[0] = *C, dst.ca[1] = C[8];
				*(uint16*)&dst.atr = *(uint16*)&spr->atr; /*atr,x*/
				*(uint32*)&SPRBUF[ns<<2] = *(uint32*)&dst;
			}
			++ns;
		} while(((uint8*)spr)+=4,--n!=0);

		if (ns>8) /* handle >8 sprites per scanline */
			PPU_status|= 0x20;
	} else
	{
		/*for(n=63;n>=0;n--,spr++)*/
		do
		{
			unsigned	v = scanline-spr->y;

			if (v>=H)
				continue;
			if (ns>=ms)
				{PPU_status|= 0x20; break;}

			if (n==64) /*==63*/
				sb = 1;

			{
				uint8		*C;
				unsigned int	vadr = spr->no;

				if (Sprite16)
					vadr = ((vadr&1)<<12) + ((vadr&0xFE)<<4);
				else
					vadr = (vadr<<4)+vofs;

				if (spr->atr&V_FLIP)
					vadr+= 7, vadr-= v, vadr+= Pf, vadr-= v&8;
				else
					vadr+= v, vadr+= v&8;

				if (MMC5Ex)
					C = MMC5SPRVRAMADR(vadr);
				else
					C = VRAMADR(vadr);
				dst.ca[0] = *C;

				if (ns<8)
				{
					PPU_hook(0x2000);
					PPU_hook(vadr);
				}

				dst.ca[1] = C[8];
				*(uint16*)&dst.atr = *(uint16*)&spr->atr; /*atr,x*/
				*(uint32*)&SPRBUF[ns<<2] = *(uint32*)&dst;
			}
			++ns;
		} while(((uint8*)spr)+=4,--n!=0);

		if (ns>8) /* handle >8 sprites per scanline */
			PPU_status|= 0x20;
		else if ((n=8-ns)>0)
			do
			{
				PPU_hook(0x2000);
				PPU_hook(vofs);
			} while(--n!=0);
	}

	PPUnumsprites = ns;
	SpriteBlurp = sb;
}

void RefreshSprites(void);
void RefreshSprites()
{
	SPRB	*spr;
	int	n;

	PPUspork = 0;
	if ((n=PPUnumsprites)==0)
		return;

	FCEU_dwmemset(PPUsprlinebuf,0x80808080,256);
	PPUnumsprites = --n;
	spr = n+(SPRB*)SPRBUF;
	do
	{
		register uint32	pixdata;
		register uint8	J = spr->ca[0], atr;

		uint8	*C, *VB;
		int	x;

		pixdata = PPU_LUT1to4[spr->ca[1]], pixdata+= pixdata;
		pixdata|= PPU_LUT1to4[J];

		if ((J|=spr->ca[1])==0)
			continue;
		x = spr->x, atr = spr->atr;
		if (n==0 && SpriteBlurp && (PPU_status&0x40)==0)
		{
			if ((uint8)(atr&H_FLIP)==0)
				PPUsphitdata = J;
			else
				PPUsphitdata = bitreverseLUT[J];
			PPUsphitx = x;
		}

		C = PPUsprlinebuf+x;
		VB = (PALRAM+0x10)+((atr&3)<<2);

		if ((uint8)(atr&SP_BACK))
			if ((uint8)(atr&H_FLIP))
			{
				if ((uint8)(J&0x80)) C[7] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[6] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[5] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[4] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[3] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[2] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[1] = VB[pixdata&3]|0x40;
				if ((J+=J)&0x80) *C = VB[pixdata>>=4]|0x40;
			} else
			{
				if ((uint8)(J&0x80)) *C = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[1] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[2] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[3] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[4] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[5] = VB[pixdata&3]|0x40;
				pixdata>>= 4;
				if ((J+=J)&0x80) C[6] = VB[pixdata&3]|0x40;
				if ((J+=J)&0x80) C[7] = VB[pixdata>>=4]|0x40;
			}
		else
			if ((uint8)(atr&H_FLIP))
			{
				if ((uint8)(J&0x80)) C[7] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[6] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[5] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[4] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[3] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[2] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[1] = VB[pixdata&3];
				if ((J+=J)&0x80) *C=VB[pixdata>>=4];
			} else
			{
				if ((uint8)(J&0x80)) *C = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[1] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[2] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[3] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[4] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[5] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80) C[6] = VB[pixdata&3];
				if ((J+=J)&0x80) C[7] = VB[pixdata>>=4];
			}
	} while(--spr,--n>=0);

	SpriteBlurp = 0;
	PPUspork = 1;
}

void CopySprites(uint8 *target);
void CopySprites(uint8 *target)
{
	const uint8	*S;
	uint8		*P;
	uint8		n;

	n = ~PPU[1], n+= n, n&= 8; /* have SpFirst8 determine where to start */
	S = n+PPUsprlinebuf;
	P = n+target;
	if (!PPUspork)
		return;
	PPUspork = 0;

	if (PPUrendis&1)
		return; /* User asked to not display sprites. */

	do
	{
		uint8 t;

		if (*(uint32*)S==0x80808080)
			{S+=4; continue;}

	/* act on each normal or behind bg sprite pixel */
		if (!((t=*S++)&0x80) && (!(t&0x40) || *P&0x40))
			*P = t;
		if (!((t=*S++)&0x80) && (!(t&0x40) || P[1]&0x40))
			P[1] = t;
		if (!((t=*S++)&0x80) && (!(t&0x40) || P[2]&0x40))
			P[2] = t;
		if (!((t=*S++)&0x80) && (!(t&0x40) || P[3]&0x40))
			P[3] = t;
	} while(P+=4,(n+=4)!=0);
}

/* line rendering */

static void DoHBlank(void)
{
	uint32	rad;

	if (!RenderON)
		return;
	rad = RefreshAddr;
	if ((~rad&0x7000)==0) /*(rad&0x7000)==0x7000*/
	{
		uint32 t;

		rad^= 0x7000;
		if ((t=~rad&0x3E0)==0) /*(rad&0x3E0)==0x3E0*/
			rad^= 0x3E0;
		else if (t==0x40) /*(rad&0x3E0)==0x3A0*/
			rad^= 0x3A0^0x800;
		else
			rad+= 0x20;
	} else
		rad+= 0x1000;
	RefreshAddr = rad;
}

static void FASTAPASS(1) RefreshLine(int lastpixel)
{
	static uint32	pshift[2];
	static uint32	atlatch;
	uint32	smorkus = RefreshAddr;
#define RefreshAddr	smorkus
	uint32	vofs;
	int	X1;

	register uint8	*P;
	int	prefetchtile = lastpixel>>3;
	int	numtiles;
	int	xdownshift = 8-XOffset;
	int	scantileline = scanline>>3;

	/* Yeah, recursion would be bad.
		PPU_hook() functions can call mirroring/chr bank switching functions,
		which call FCEUPPU_LineUpdate, which call this function.
	*/
	static int norecurse = 0;
	if (norecurse)
		return;

	if (PPUsphitx!=0x100 && (PPU_status&0x40)==0)
		if (PPUsphitx<(lastpixel-16) && PPUsphitx>=(prefetchtile-2)*8)
			++prefetchtile;

	if (prefetchtile>34)
		prefetchtile = 34;
	if ((numtiles=prefetchtile-firsttile)<=0)
		return;

	P = PPU_Pline;
	vofs = (BGAdrHI<<8)|(RefreshAddr>>12&7);

	if (!RenderON)
	{ //TODO: to get the pretty colors effect to work, CPU execution here
		uint32	tem = PALRAM[0]|0x40;

		tem|= tem<<8, tem|= tem<<16;
		FCEU_dwmemset(PPU_Pline,tem,numtiles*8);
		P+= numtiles*8;
		PPU_Pline = P;
		firsttile = prefetchtile;

#define TOFIXNUM (272-0x4)
		if (lastpixel>=TOFIXNUM && endofline)
			{DoHBlank(); endofline = 0;}

		if (InputScanlineHook && (lastpixel-16)>=0)
			InputScanlineHook(PPU_Plinef,
				PPUspork ? PPUsprlinebuf:0,
				linestartts,
				prefetchtile*8-16);
		return;
	}

	/* Priority bits, needed for sprite emulation. */
	PALRAM[0x0]|= 0x40;
	PALRAM[0x4]|= 0x40;
	PALRAM[0x8]|= 0x40;
	PALRAM[0xC]|= 0x40;

	if (PPUrendis&4) /* skip rendering of background (yes, some code put here) */
	{ /* this doesn't work for all ROMs (expect some to crash) */
		uint32	me32 = ppu_void_color!=0xFF ?
			ppu_void_color:PALRAM[4]|0x80; //substitute void color

		me32|= me32<<8, me32|= me32<<16;
		for(X1=firsttile; X1<prefetchtile; ++X1)
		{ /* fake everytin' */
			if (PPU_hook)
			{
				PPU_hook(0x2000|(RefreshAddr&0xFFF));
				PPU_hook(vofs);
			}

			*((uint32*)P)++ = me32, *((uint32*)P)++ = me32;

			if ((~RefreshAddr&HT_POS_MASK)==0)
				RefreshAddr^= H_POS_MASKS;
			else
				++RefreshAddr;

			if (PPU_hook)
				PPU_hook(0x2000|(RefreshAddr&0xFFF));
		}
		atlatch = 0;
		goto skip;
	}

	/* This high-level graphics MMC5 emulation code was written
		for MMC5 carts in "CL" mode.  It's probably not totally
		correct for carts in "SL" mode.
	*/
#define PPUT_MMC5
	if (MMC5Ex && GenieStage!=1)
	{
		if (MMC5ExMode==0 && MMC5SplitCtrl&0x80)
		{
			int	tochange = MMC5SplitCtrl&0x1F;

			tochange-= firsttile;
			for(X1=firsttile; X1<prefetchtile; ++X1)
				if ((--tochange<0 && MMC5SplitCtrl&0x40)
						|| (tochange>=0 && !(MMC5SplitCtrl&0x40)))
				{
				#define PPUT_MMC5SP
				#include "PPUtile.h"
				#undef PPUT_MMC5SP
				} else
				{
				#include "PPUtile.h"
				}
		} else if (MMC5ExMode==1 && MMC5SplitCtrl&0x80)
		{
		//	int tochange = MMC5SplitCtrl&0x1F; /* not used */

		//	tochange-= firsttile;
			#define PPUT_MMC5SP
			#define PPUT_MMC5CHR1
			for(X1=firsttile; X1<prefetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
			#undef PPUT_MMC5CHR1
			#undef PPUT_MMC5SP
		} else if (MMC5ExMode==1)
		{
			#define PPUT_MMC5CHR1
			for(X1=firsttile; X1<prefetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
			#undef PPUT_MMC5CHR1
		} else
			for(X1=firsttile; X1<prefetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
	#undef PPUT_MMC5
	} else if (PPU_hook)
	{
		norecurse = 1;
		#define PPUT_HOOK
		for(X1=firsttile; X1<prefetchtile; ++X1)
		{
		#include "PPUtile.h"
		}
		#undef PPUT_HOOK
		norecurse = 0;
	} else
		for(X1=firsttile; X1<prefetchtile; ++X1)
		{
		#include "PPUtile.h"
		}
#undef vofs
#undef RefreshAddr
skip:
	/* Reverse changes made before. */
	PALRAM[0x0]&= 0x3F;
	PALRAM[0x4]&= 0x3F;
	PALRAM[0x8]&= 0x3F;
	PALRAM[0xC]&= 0x3F;

	RefreshAddr = smorkus;
	if (firsttile<=2 && 2<prefetchtile && !BGFirst8)
	{
		uint32	tem = PALRAM[0]|0x40;

		tem|= tem<<8, tem|= tem<<16;
		*(uint32*)PPU_Plinef = *(uint32*)(PPU_Plinef+4) = tem;
	}

	if (!ScreenON)
	{
		int	tstart, tcount;
		uint32	tem = PALRAM[0]|0x40;

		tem|= tem<<8, tem|= tem<<16;
		tcount = prefetchtile-firsttile;
		if ((tstart=firsttile-2)<0)
			tcount+= tstart, tstart = 0;
		if ((tcount*=8)>0)
			FCEU_dwmemset(PPU_Plinef+tstart*8,tem,tcount);
	}

	if (lastpixel>=TOFIXNUM && endofline)
	{
		//puts("Fixed");
		DoHBlank();
		endofline = 0;
	}

	//CheckSpriteHit(lasttile*8); //lasttile*8); //lastpixel);

	/* This only works right because of a hack earlier in this function. */
	CheckSpriteHit(lastpixel);

	if (InputScanlineHook && (lastpixel-16)>=0)
		InputScanlineHook(PPU_Plinef,
			PPUspork ? PPUsprlinebuf:0,
			linestartts,
			prefetchtile*8-16);

	PPU_Pline = P;
	firsttile = prefetchtile;
}

static void line32_or(uint32 *d, uint32 k)
{
	int		x = 8;

	do
		*d++|= k, *d++|= k, *d++|= k, *d++|= k,
		*d++|= k, *d++|= k, *d++|= k, *d++|= k;
	while(--x!=0);
}
static void line32_and(uint32 *d, uint32 m)
{
	int		x = 8;

	do
		*d++&= m, *d++&= m, *d++&= m, *d++&= m,
		*d++&= m, *d++&= m, *d++&= m, *d++&= m;
	while(--x!=0);
}
static void line32_mask(uint32 *d, uint32 m, uint32 k)
{
	int		x = 8;

	do
		*d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k,
		*d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k;
	while(--x!=0);
}

#define DoUpdateV() do{ \
	if(RenderON) \
		RefreshAddr=(RefreshAddr&0xFBE0)|(TempAddr&0x41F); \
}while(0)

static void DoLine(void)
{
	uint8	*target = XBuf+(scanline<<8);

	if (MMC5Ex && RenderON)
		MMC5_hb(scanline);

	X6502_Run(256);

	/* EndRL */
	RefreshLine(272);
	if (endofline)
		DoHBlank();
	CheckSpriteHit(272);
	PPU_Pline = 0;

	if (PPUrendis&2) /* User asked to not display background data. */
	{
		uint32	tem = ppu_void_color!=0xFF ?
			ppu_void_color:PALRAM[0]|0x40; //substitute void color

		tem|= tem<<8, tem|= tem<<16;
		FCEU_dwmemset(target,tem,256);
	}

	if (SpriteON)
		CopySprites(target);

	if (RenderON) /* Yes, very el-cheapo. */
		if (GrayLine)
			line32_and((uint32*)target,0x30303030);

	if ((~PPU[1]&Deempbits)==0) /*(PPU[1]>>5)==0x7*/
		line32_mask((uint32*)target,0x3F3F3F3F,0xC0C0C0C0);
	else if (Deemphas)
		line32_or((uint32*)target,0x40404040);
	else
		line32_mask((uint32*)target,0x3F3F3F3F,0x80808080);

	PPUsphitx = 0x100;

	if (RenderON)
		FetchSpriteData();

	if (GameHBIRQHook && RenderON && (PPU[0]&0x38)!=0x18)
	{
		X6502_Run(6);
		DoUpdateV();

		X6502_Run(4);
		GameHBIRQHook();
		X6502_Run(85-16-10);
	} else
	{
		X6502_Run(6);
		DoUpdateV();

		X6502_Run(85-6-16);

		// A semi-hack for Star Trek: 25th Anniversary
		if (GameHBIRQHook && RenderON && ((PPU[0]&0x38)!=0x18))
			GameHBIRQHook();
	}

	if (SpriteON)
		RefreshSprites();
	if (GameHBIRQHook2 && RenderON)
		GameHBIRQHook2();
}

void FCEUPPU_SetVideoSystem(int which)
{
	if (which==1)
	{ /* PAL */
		ppuvbl_scanline_cc = FSettings.UsrVBLSLines[1]*LCYCS; /*312-242*/
		FSettings.FirstSLine = FSettings.UsrFirstSLine[1];
		FSettings.LastSLine = FSettings.UsrLastSLine[1];
	} else if (which==0)
	{ /* NTSC */
		ppuvbl_scanline_cc = FSettings.UsrVBLSLines[0]*LCYCS; /*262-242*/
		FSettings.FirstSLine = FSettings.UsrFirstSLine[0];
		FSettings.LastSLine = FSettings.UsrLastSLine[0];
	}

	if ((signed)ppuvbl_scanline_cc<FIRSTCYC) /* idiot proofing */
		ppuvbl_scanline_cc = 0;
	else
		ppuvbl_scanline_cc-= FIRSTCYC;
}

void FCEUPPU_Init(void)
{
	makeluts(PPU_LUT1to4,PPU_TileLUT,bitreverseLUT);
}

void FCEUPPU_Reset(void)
{
	VRAMBuffer = PPU[0] = PPU[1] = PPU_status = PPU[3] = 0;
	PPUSPL = 0;
	PPUGenLatch = 0;
	RefreshAddr = TempAddr = 0;
	vtoggle = 0;
	PPUBoot = 2;
	DummyLineEndCC = 15;
}

void FCEUPPU_Power(void)
{
	readfunc	*rdrt;
	writefunc	*wrrt;
	int		x;

	UPALRAM[0x4] = UPALRAM[0x8] = UPALRAM[0xC] = 0;
	FCEU_memsetl(NTARAM,0,0x800);
	FCEU_memsetl(PALRAM,0,0x20);
	FCEU_memsetl(SPRAM,0,0x100);
	FCEUPPU_Reset();

	x = 0x2000, rdrt = x+ARead, wrrt = x+BWrite;
	do
		*rdrt++ = PPUA200x,
		*rdrt++ = PPUA200x,
		*rdrt++ = PPUA2002,
		*rdrt++ = PPUA200x,
		*rdrt++ = PPUA200x, //A2004;
		*rdrt++ = PPUA200x,
		*rdrt++ = PPUA200x,
		*rdrt++ = PPUA2007,

		*wrrt++ = PPUB2000,
		*wrrt++ = PPUB2001,
		*wrrt++ = PPUB2002,
		*wrrt++ = PPUB2003,
		*wrrt++ = PPUB2004,
		*wrrt++ = PPUB2005,
		*wrrt++ = PPUB2006,
		*wrrt++ = PPUB2007;
	while((x+=8)<0x4000);
	BWrite[0x4014] = PPUB4014;
}

static void ResetRL(uint8 *target)
{
	FCEU_memsetl(target,-1,256);
	if (InputScanlineHook)
		InputScanlineHook(0,0,0,0);
	PPU_Plinef = target;
	PPU_Pline = target;
	firsttile = 0;
	linestartts = timestamp*48+XFrame.count;
	endofline = 0;
	FCEUPPU_LineUpdate();
	endofline = 1;
}

/* PPU-oriented frame */

void FCEUPPU_Loop(void)
{
	if (PPUBoot) //Needed for Knight Rider, possibly others.
	{
		FCEU_memsetl(XBuf,0x80U*0x01010101,256*240);
		if (--PPUBoot!=0)
		{
			scanline = 241;
			X6502_Run(ppuvbl_scanline_cc+FIRSTCYC);
		} else
		{
			scanline = 0;
			X6502_Run(242*LCYCS);
		}
		goto rendered;
	}

	PPU_status|= 0x80; //vertical interrupt phase
//	scanline = 241; //added; 2009.12.29

	// reset sprites
	//Not having this here breaks a Super Donkey Kong game.
	PPU[3] = PPUSPL = 0; 
	//I need to figure out the true nature and length of this delay.
	X6502_Run(FIRSTCYC);

	if (FCEUGameInfo->type==GIT_NSF)
		DoNSFFrame();
	else if (VBlankON)
		XFrame.IRQlow|= FCEU_IQNMI; //TriggerNMI

	X6502_Run(ppuvbl_scanline_cc);

	PPU_status&= 0x1F; //out of VINT; onto dummy line
	X6502_Run(256);

	if (RenderON)
	{
		if (GameHBIRQHook && (PPU[0]&0x38)!=0x18)
			GameHBIRQHook();
		if (PPU_hook)
		{
			int	x = 42;

			do
			{
				PPU_hook(0x2000);
				PPU_hook(0);
			} while(--x!=0);
		}

		if (GameHBIRQHook2)
			GameHBIRQHook2();
	}

	X6502_Run(85-16);

	if (RenderON)
	{
		RefreshAddr = TempAddr;
		if (PPU_hook)
			PPU_hook(RefreshAddr&0x3fff);
	}

	//Clean this stuff up later.
	PPUspork = PPUnumsprites = 0;
	ResetRL(XBuf);

	if (!PAL) //only on NTSC; 2010.03.17
		DummyLineEndCC^= 15^16;
	X6502_Run(DummyLineEndCC); //16-kook);

	if (FCEUGameInfo->type==GIT_NSF)
		X6502_Run(LCYCS*240);
	//frame skipping no longer supported (sorry)
#ifdef FRAMESKIP
/*	else if (skip)
	{
		int	y = 1+SPRAM[0];

		PPU_status|= 0x20;       // Fixes "Bee 52".  Does it break anything?
		if (GameHBIRQHook)
		{
			X6502_Run(256);
			for(scanline=0; scanline<240; ++scanline)
			{
				if (RenderON)
					GameHBIRQHook();
				if (scanline==y && SpriteON)
					PPU_status|= 0x40;
				X6502_Run(scanline==239 ? 85:LCYCS);
			}
		} else if (y<240)
		{
			X6502_Run(LCYCS*y);
			if (SpriteON)
				PPU_status|= 0x40; // Quick and very dirty hack.
			X6502_Run(LCYCS*(240-y));
		} else
			X6502_Run(LCYCS*240);
	}*/
#endif
	else
	{
		static int	deempcnt[8];

		PPUemp = PPU[1]>>5;
		scanline = 0;
		do
		{
			++deempcnt[PPUemp];
			DoLine();
			if (++scanline<240)
				ResetRL(XBuf+(scanline<<8));
			X6502_Run(16);
		} while(scanline<240);

		if (MMC5Ex && RenderON)
			MMC5_hb(scanline);

		{
			int	*deempcntp = 1+deempcnt;
			int	x = 1, max = 0, maxref = 0;

			do
			{
				if (*deempcntp>max)
					max = *deempcntp, maxref = x;
				*deempcntp++ = 0;
			} while(++x<7);

			SetNESDeemph(maxref,0);
		}
	}
	//this prev. called at beginning; 2010.03.17
	X6502_Run(LCYCS);
rendered:
	FCEU_PutImage();
}

static uint16 TempAddrT, RefreshAddrT;
static uint8 kookT;

void FCEUPPU_LoadState(int v)
{
(void)v;
	TempAddr = TempAddrT;
	RefreshAddr = RefreshAddrT;
	DummyLineEndCC = kookT+15;
}

SFORMAT FCEUPPU_STATEINFO[] =
{
  { NTARAM, 0x800, "NTAR"},
  { PALRAM, 0x20, "PRAM"},
  { 1+UPALRAM, 12, "UPAL"}, //2010.03.08
  { SPRAM, 0x100, "SPRA"},
  { PPU, 0x4, "PPUR"},
  { &kookT, 1, "KOOK"},
  { &PPUBoot, 1, "DEAD"},
  { &PPUSPL, 1, "PSPL"},
  { &XOffset, 1, "XOFF"},
  { &vtoggle, 1, "VTOG"},
  { &RefreshAddrT, 2|FCEUSTATE_RLSB, "RADD"},
  { &TempAddrT, 2|FCEUSTATE_RLSB, "TADD"},
  { &VRAMBuffer, 1, "VBUF"},
  { &PPUGenLatch, 1, "PGEN"},
  { 0 }
 };

void FCEUPPU_SaveState(void)
{
	TempAddrT = TempAddr;
	RefreshAddrT = RefreshAddr;
	kookT = DummyLineEndCC-15;
}
