/*
 boards/unlicensed.c
  short unlicensed, also bootleg boards

 * MacFCEU 0.5.5
 *
 * Copyright notice for this file:
 *  mmc3.cpp Copyright (C) 1998 BERO
 *  mmc3.cpp Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"
#include "MMC3and6.h"
#include "unlicensed.h"

/* ---- BMC Golden 190-in-1 */

static void BMC190in1Sync(void)
{
	int		v = latch_byte;

	setmirror(~v&1);
	setprg16(0x8000,v>>=2);
	setprg32(0xC000,v);
	setchr8(v);
}

void BMC_190in1_Init(CartInfo *info)
{
	Latch_Init(info,BMC190in1Sync,0,0,0x8000FFFFU);
}

/* ---- UNL-CC21 */

static void UNLCC21Sync(void)
{
	setprg32(0x8000,0);
	setchr8(latch_addr);
	setmirror(MI_0+(latch_addr>>1&1));
}

void UNL_CC21_Init(CartInfo *info)
{
	Latch_Init(info,UNLCC21Sync,0,0x8000,0x8000FFFFU);
}

/* ---- mapper 44: BMC SuperBig 7-in-1 */

static void M44PW(uint32 A, uint8 V)
{
	uint32	NV = V;

	if (MMC3EXPREGS[0]<6)
		NV&= 0xF;
	else
		NV&= 0x1F;
	NV|= MMC3EXPREGS[0]<<4;
	setprg8(A,NV);
}

static void M44CW(uint32 A, uint8 V)
{
	uint32	NV = V;

	if (MMC3EXPREGS[0]<6)
		NV&= 0x7F;
	NV|= MMC3EXPREGS[0]<<7;
	setchr1(A,NV);
}

static DECLFW(M44_write)
{
(void)A;
	V&= 7;
	if (MMC3EXPREGS[0]==V)
		return;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M44Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandlerI(2,0xA001,0xBFFF,M44_write);
}

void BMC_SuperBig7in1_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	info->Power = M44Power;
	MMC3cwrap = M44CW;
	MMC3pwrap = M44PW;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 52: BMC Mario Party 7-in-1 */

static void M52PW(uint32 A, uint8 V)
{
	uint32	NV = V;

	NV&= 0x1F^(MMC3EXPREGS[0]<<1&0x10);
	NV|= ((MMC3EXPREGS[0]&6)|(MMC3EXPREGS[0]>>3&MMC3EXPREGS[0]&1))<<4;
	setprg8(A,NV);
}

static void M52CW(uint32 A, uint8 V)
{
	uint32	NV = V;

	NV&= ~(MMC3EXPREGS[0]<<1&0x80);
	NV|= ((MMC3EXPREGS[0]>>3&4)|(MMC3EXPREGS[0]>>1&2)
		|(MMC3EXPREGS[0]>>6&MMC3EXPREGS[0]>>4&1))<<7;
	setchr1(A,NV);
}

static DECLFW(M52_write)
{
	if (MMC3EXPREGS[1]) //update once, then set latch
	{
		MMC3WRAM[A-0x6000] = V;
		return;
	}
	MMC3EXPREGS[1] = 1;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M52Power(void)
{
	MMC3EXPREGS[0] = MMC3EXPREGS[1] = 0;
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M52_write);
}

void BMC_MarioParty7in1_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	info->Power = M52Power;
	MMC3cwrap = M52CW;
	MMC3pwrap = M52PW;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}

/* ---- mapper 254: bootleg Pikachu Y2K */
/* there's also an Ai Senshi Nicol pirate, but piracy is illegal... */

static DECLFR(M254_read)
{
	//if (MMC3EXPREGS[0])
	//	return MMC3WRAM[A-0x6000];
	//else
	return MMC3WRAM[A-0x6000]^MMC3EXPREGS[1];
}

static DECLFW(M254_write)
{
	MMC3EXPREGS[0] = 0;
//	if (A==0x8000)
//		MMC3EXPREGS[0] = 0xFF;
//	else if (A==0xA001)
//		MMC3EXPREGS[1] = V;
	MMC3_CMDWrite(A,V);
}

static void M254Reset(void)
{
	MMC3EXPREGS[0] = ~0;
	GenMMC3Reset();
}

static void M254Power(void)
{
	MMC3EXPREGS[0] = ~0;
	GenMMC3Power();
	SetReadHandler(0x6000,0x7FFF,M254_read);
	SetWriteHandlerI(2,0x8000,0x9FFE,M254_write);
}

void BTL_PikachuY2K_Init(CartInfo *info)
{
	GenMMC3_Init(info,128,128,8,info->battery);
	info->Power = M254Power;
	info->Reset = M254Reset;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}
