/*
 enviro64.h
  Bigger-than-reason environment include,
   featuring 32 bit macros inlines for 64 bit routines;
    so good, mycseries is effectively dead.

  Created: 2008.06.16
  Revised: 2010.01.17

Tabs are set to 8.
Window is adjusted to 90 char/line.

In use:
  `int` should be 32 bits (4 bytes), and it goes without saying.
  `long int` should be at least as large as `void*`.
  The former xintxx_t here are now just xintxx (remove interfering typedefs).

Values:
  The 64 bit routines do not clip the result; they may return the lower or higher 32
bits and/or sign depending on the application.

Warnings:
  Some of the macros under native 64 bit may call for a (void) cast.
	e.g., `(void)int64addu32(v,l)`

IMPORTANT:
  A number of the macros may use any of their parameters more than once; it's recommended
that you use only literals or direct variables for those.

  Let the compiler deal with optimizations.

  Routines/macros that use their parameters only _once_:
	int64hi
	int64lo
	int64n
	int64addhl
	uint64addhl
	int64addhls
	uint64addhls
	int64addu32
	uint64addu32
	int64subu32
	uint64subu32
	muls64s32
	muls64u32
	mulu64s32
	mulu64u32
	int64muls32
	uint64muls32
	int64mulu32
	uint64mulu32
	int64divu32
	uint64divu32
*/

#ifndef __ENVIRO64_H
/*
  avoid casts at all costs!
*/

typedef char			int8;
typedef short			int16;
#if sizeof(long)==4
typedef long			int32;
#else
typedef int			int32;
#endif

typedef unsigned char		uint8;
typedef unsigned short		uint16;
#if sizeof(long)==4
typedef unsigned long		uint32;
#else
typedef unsigned int		uint32;
#endif

/* most of this part is from ConditionalMacros.h
 (not explicitly #include'd for performance reasons)
*/

/* enivonment define's:
	env68k			MC68020 compatible (Motorola, Inc.)
	envppc			PowerPC 60x (Motorola/IBM)
	envx86			80386 compatible (Intel Corp.)
	envmrx			MIPS MRX000 (Silicon Graphics Inc.)
	envspr			Sparc (Sun Microsystems Inc.)
*/

#if defined(__MRC__)
 #define envppc 1
 #define LONG_LONG 1
 #if (__MRC__>0x0300) && (__MRC__<0x0700)
  #if __option(longlong)
   #define LONG_LONG 1
  #endif
 #endif

#elif defined(__SC__) && (defined(MPW_CPLUS) || defined(MPW_C))
 #if __option(mc68020)
  #define env68k 1
 #endif

#elif defined(__MWERKS__)
 #if (__MWERKS__<0x0900) || macintosh
  #if powerc
   #define envppc 1
  #else
   #define env68k 1
  #endif
 #elif ((__MWERKS__>=0x1900) && __MIPS__) || ((__MWERKS__>=0x2110) && __MACH__)
  #if __MIPS__
   #define envmrx 1
  #elif __MACH__
   #define envppc 1
  #endif

  #if __option(little_endian)
   #define BIG_ENDIAN 0
  #else
   #define BIG_ENDIAN 1
  #endif
 #endif

 #if (__MWERKS__>=0x1100)
  #if __option(longlong)
   #define LONGLONG
  #endif
 #endif

#elif defined(SC_PLUS_THINK) || defined(SYMANTEC_CPLUS) || defined(SYMANTEC_C)
 #if powerc
  #define envppc 1
 #else
  #define env68k 1
 #endif

#elif defined(__GNUC__) && (defined(__APPLE_CPP__) || defined(__APPLE_CC__)
	|| defined(__NEXT_CPP__))
 #if defined(__ppc__) || defined(powerpc) || defined(ppc)
  #define envppc 1
 #elif defined(m68k)
  #define env68k 1
 #elif defined(sparc)
  #define envspr 1
 #elif defined(__i386__) || defined(i386) || defined(intel)
  #define envx86 1
  #define BIG_ENDIAN 0
 #endif

 #if __GNUC__>=2
  #define LONG_LONG 1
 #endif

#elif defined(__GNUC__) && defined(__MINGW32__)
 #define envx86 1
 #define LONG_LONG 1

#elif defined(__GNUC__)
 #define LONG_LONG 1
 #if #cpu(powerpc)
  #define envppc 1
 #elif #cpu(m68k)
  #define env68k 1
 #endif

 #if defined(_LONG_LONG)
  #define LONG_LONG 1
 #else
  #define LONG_LONG 0
 #endif

#elif defined(__xlc) || defined(__xlC) || defined(__xlC__)
	|| defined(__XLC121__)
 #define envppc 1
 #if defined(_LONG_LONG)
  #define LONG_LONG 1
 #else
  #define LONG_LONG 0
 #endif

#elif defined(_MSC_VER) && !defined(__MWERKS__) 
 #if defined(_M_M68K)
  #define env68k 1
 #elif defined(_M_MPPC) || defined(_M_PPC)
  #define envppc 1 /* _LONG_LONG not supported */
 #elif defined(_M_IX86)
  #define envx86 1
 #elif defined(_M_ALPHA)
  #define envalp 1
 #elif defined(_M_MRX000)
  #define envmrx 1
 #endif

#elif defined(__MOTO__)
 #define envppc 1 /* no known way to detect _LONG_LONG */

#elif defined(_MIPS_ISA)
 #define envmrx 1

#elif defined(__sparc)
 #define envspr 1

#elif MSVC
 #define LONG_LONG 1

 /* warning: THINK C/Symantec C++ and MPWC/PPCC are obsolete
   and cannot fully handle the C language
 */
#elif defined(THINK_C) || (defined(applec) && !defined(__SC__))
 #define env68k 1

#elif defined(__PPCC__)
 #define envppc 1

#else
 #error Can't identify the compiler, environment.
#endif

 /* default little endian */
#if !defined(BIG_ENDIAN)
 #if defined(envppc) || defined(env68k) || defined(envspr)
  #define BIG_ENDIAN 1
 #else
  #define BIG_ENDIAN 0
 #endif
#endif

 /* default no long long */
#if !defined(LONG_LONG)
 #define LONG_LONG 0
#endif

 /* long long typedef, macros, inlines for environs. that don't support them */

#if LONG_LONG
 #if MSVC
  typedef __int64		int64;
  typedef unsigned __int64	uint64;
 #else
  typedef long long		int64;
  typedef unsigned long long	uint64;
 #endif

 #define _int64val(h,l)		(((int64)(h)<<32)+(uint32)(l))
 #define _int64xval(f)		((int64)(f))

 #define _int64xset(x,f)	(void)((x)=(long double)(f))
 #define int64xset(x,f)		(void)((x)=(long double)(f))
 #define int64sethl(x,h,l)	(void)((x)=((int64)(h)<<32)+(uint32)(l))
 #define int64setl(x,y)		(void)(x,(x)=(long int)(y))
 #define int64set(x,y)		(void)(x,y,(x)=y)

 #define int64hi(x)		((x)>>32)
 #define int64lo(x)		(uint32)(x)

 #define int64z(n)		(n,n==0)
 #define int64n(n)		((int64)(n)<0)

 #define int64toi(n)		((int32)(n)==(int64)(n)?(int32)(n):0x7FFFFFFFL+(n>>64))

 #define int64eq(x,y)		(x,y,(x)==(y))
 #define int64eqhl(x,h,l)	(x,(x)==((int64)(h)<<32)+(uint32)(l))
 #define int64gt(x,y)		(x,y,(x)>(y))
 #define int64gthl(x,h,l)	(x,h,(x)>((int64)(h)<<32)+(uint32)(l))
 #define int64ge(x,y)		(x,y,(x)>=(y))
 #define int64gehl(x,h,l)	(x,h,(x)>=((int64)(h)<<32)+(uint32)(l))
 #define uint64gtu32(x,i)	(x,(uint64)(x)>(uint32)(i))

 #define int64addhl(x,h,l)	(((x)+=((int64)(h)<<32)+(uint32)(l))>>32)
 #define uint64addhl(x,h,l)	(((x)+=((int64)(h)<<32)+(uint32)(l))>>32)

 #define int64addhls(x,h,l)	(int64)((x)+=((int64)(h)<<32)+(uint32)(l))
 #define uint64addhls(x,h,l)	(uint64)((x)+=((int64)(h)<<32)+(uint32)(l))

 #define int64addu32(x,l)	(((x)+=(uint32)(l))>>32)
 #define uint64addu32(x,l)	(((x)+=(uint32)(l))>>32)
 #define int64add(x,y)		(((x)+=(y))>>32)
 #define uint64add(x,y)		(((x)+=(y))>>32)

 #define int64subhl(x,h,l)	(((x)-=((int64)(h)<<32)+(uint32)(l))>>32)
 #define uint64subhl(x,h,l)	(((x)-=((int64)(h)<<32)+(uint32)(l))>>32)

 #define int64subu32(x,l)	(((x)-=(uint32)(l))>>32)
 #define uint64subu32(x,l)	(((x)-=(uint32)(l))>>32)
 #define int64sub(x,y)		(((x)-=(y))>>32)
 #define uint64sub(x,y)		(((x)-=(y))>>32)

 /* these seven binary macros added; 2010.01.17 */
 #define int64or(x,y)		(uint32)(x,y,(x)|=(y))
 #define int64eor(x,y)		(uint32)(x,y,(x)^=(y))
 #define int64and(x,y)		(uint32)(x,y,(x)&=(y))
 #define int64not(x)		(uint32)(x,x,(x)=~(x))

 #define int64orhl(x,h,l)	(uint32)(x,((x)|=((int64)(h)<<32)+(uint32)(l))>>32)
 #define int64eorhl(x,h,l)	(uint32)(x,((x)^=((int64)(h)<<32)+(uint32)(l))>>32)
 #define int64andhl(x,h,l)	(uint32)(x,((x)^=((int64)(h)<<32)+(uint32)(l))>>32)

 #define _int64val(h,l)		(((int64)(h)<<32)+(uint32)(l))
 #define _int64xval(f)		((int64)(f))

 #define _int64xset(x,f)	(void)((x)=(long double)(f))
 #define int64xset(x,f)		(void)((x)=(long double)(f);}while(0)
 #define int64sethl(x,h,l)	(void)((x)=((int64)(h)<<32)+(uint32)(l))
 #define int64setl(x,y)		(void)(x,(x)=(long int)(y))
 #define int64set(x,y)		(void)(x,y,(x)=y)

 #define int64hi(x)		((x)>>32)
 #define int64lo(x)		(uint32)(x)

 #define int64z(n)		(n,n==0)
 #define int64n(n)		((int64)(n)<0)

 #define int64toi(n)		((int32)(n)==(int64)(n)?(int32)(n):0x7FFFFFFFL+(n>>64))

 #define int64eq(x,y)		(x,y,(x)==(y))
 #define int64eqhl(x,h,l)	(x,(x)==((int64)(h)<<32)+(uint32)(l))
 #define int64gt(x,y)		(x,y,(x)>(y))
 #define int64gthl(x,h,l)	(x,h,(x)>((int64)(h)<<32)+(uint32)(l))
 #define int64ge(x,y)		(x,y,(x)>=(y))
 #define int64gehl(x,h,l)	(x,h,(x)>=((int64)(h)<<32)+(uint32)(l))
 #define uint64gtu32(x,i)	(x,(uint64)(x)>(uint32)(i))

 #define int64addhl(x,h,l)	(((x)+=((int64)(h)<<32)+(uint32)(l))>>32)
 #define uint64addhl(x,h,l)	(((x)+=((int64)(h)<<32)+(uint32)(l))>>32)

 #define int64addhls(x,h,l)	(int64)((x)+=((int64)(h)<<32)+(uint32)(l))
 #define uint64addhls(x,h,l)	(uint64)((x)+=((int64)(h)<<32)+(uint32)(l))

 #define int64addu32(x,l)	(((x)+=(uint32)(l))>>32)
 #define uint64addu32(x,l)	(((x)+=(uint32)(l))>>32)
 #define int64add(x,y)		(((x)+=(y))>>32)
 #define uint64add(x,y)		(((x)+=(y))>>32)

 #define int64subhl(x,h,l)	(((x)-=((int64)(h)<<32)+(uint32)(l))>>32)
 #define uint64subhl(x,h,l)	(((x)-=((int64)(h)<<32)+(uint32)(l))>>32)

 #define int64subu32(x,l)	(((x)-=(uint32)(l))>>32)
 #define uint64subu32(x,l)	(((x)-=(uint32)(l))>>32)
 #define int64sub(x,y)		(((x)-=(y))>>32)
 #define uint64sub(x,y)		(((x)-=(y))>>32)

 #define muls64s32(x,i,j)	(int32)((x)=(int32)(i)*(int32)(j))
 #define muls64u32(x,i,j)	(int32)((x)=(uint32)(i)*(uint32)(j))
 #define mulu64s32(x,i,j)	(uint32)((x)=(int32)(i)*(int32)(j))
 #define mulu64u32(x,i,j)	(uint32)((x)=(uint32)(i)*(uint32)(j))

 #define int64muls32(x,l)	(int32)(((int64)(x))*=(int32)(l))
 #define uint64muls32(x,l)	(uint32)(((uint64)(x))*=(int32)(l))
 #define int64mulu32(x,l)	(int32)(((uint64)(x))*=(uint32)(l))
 #define uint64mulu32(x,l)	(uint32)(((uint64)(x))*=(uint32)(l))

 #define int64divu32(x,i)	(int32)(((int64)(x))/=(uint32)(i))
 #define uint64divu32(x,i)	(uint32)(((uint64)(x))/=(uint32)(i))

 #define int64divs32(x,i)	(int32)((i),(x)=(int64)(x)/(int32)(i))
 #define uint64divs32(x,i)	(int32)((i),(x)=(uint64)(x)/(int32)(i))

 #define imuldiv(n,m,d)		(((int64)((int32)(n))*(int32)(m))/(int32)(d))
 #define umuldiv(n,m,d)		(((uint64)((uint32)(n))*(uint32)(m))/(uint32)(d))

 #define divi32(x,i)		(int32)((int64)(x)/(int32)(i))
 #define divu32(x,i)		(uint32)((uint64)(x)/(uint32)(i))

 #define _int64shr32(x,c)	(int32)(x,(int64)(x)>>(c)>32?64:c)
 #define _uint64shr32(x,c)	(uint32)(x,(uint64)(x)>>(c)>32?64:c)

 #define int64shl(x,c)		(void)((x)=(int64)(x)<<c)
 #define int64shr(x,c)		(void)((x)=(int64)(x)>>c)
 #define uint64shr(x,c)		(void)((x)=(uint64)(x)>>c)

 #define int64half(x)		(void)(((int64)(x))/=2)
 #define uint64half(x)		(void)(((uint64)(x))>>=1)
#else
 #if BIG_ENDIAN
  typedef struct int64{int32 hi; uint32 lo;} int64;
  typedef struct uint64{uint32 hi, lo;} uint64;

  #if defined(env68k)
   /* returns high 32 bits */
   #pragma parameter __D0 _int64addhl(__A0,__D1,__D2)
    extern int32 _int64addhl(int64*,int32,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	ADD.L	D2,(A0)	*/	0xD590,
/*	ADDX.L	D1,D0	*/	0xD181,
/*	MOVE.L	D0,-(A0) */	0x2100
   };
   #pragma parameter __D0 _uint64addhl(__A0,__D1,__D2)
    extern uint32 _uint64addhl(uint64*,int32,uint32) = {
	0x2018,
	0xD590,
	0xD181,
	0x2100
    };
   /* returns high 32 bits, OR low 16 with low 8 = lo==0 ? 00:FF */
   #pragma parameter __D0 _int64addhls(__A0,__D1,__D2)
    extern int32 _int64addhls(int64*,int32,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	ADD.L	D2,(A0)	*/	0xD590,
/*	SNE	D2	*/	0x56C2,
/*	ADDX.L	D1,D0	*/	0xD181,
/*	MOVE.L	D0,-(A0) */	0x2100,
/*	OR.W	D2,D0	*/	0x8042
    };
   #pragma parameter __D0 _uint64addhls(__A0,__D1,__D2)
    extern uint32 _uint64addhls(uint64*,int32,uint32) = {
	0x2018,
	0xD590,
	0x56C2,
	0xD181,
	0x2100,
	0x8042
    };

   /* returns high 32 bits */
   #pragma parameter __D0 _int64subhl(__A0,__D1,__D2)
    extern int32 _int64subhl(int64*,int32,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	SUB.L	D2,(A0)	*/	0x9590,
/*	SUBX.L	D1,D0	*/	0x9181,
/*	MOVE.L	D0,-(A0) */	0x2100
    };
   #pragma parameter __D0 _uint64subhl(__A0,__D1,__D2)
    extern uint32 _uint64subhl(uint64*,int32,uint32) = {
	0x2018,
	0x9590,
	0x9181,
	0x2100
    };

   /* returns low 32 bits */
   #pragma parameter __D1 _int64divu32(__A0,__D2)
    extern int32 _int64divu32(int64*,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	BPL.S	*+6	;@1 */	0x6A04,
/*	NEG.L	(A0)	*/	0x4490,
/*	NEGX.L	D0	*/	0x4080,
/*@1	MOVE.L	(A0),D1	*/	0x2210,
/*	MOVE.L	D2,(A0)	*/	0x2082,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A0),D2:D1 */	0x4C50, 0x1402,
/*	TST.L	-(A0)	*/	0x4AA0,
/*	BPL.S	*+6	;@2 */	0x6A04,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@2	MOVE.L	D0,(A0)+ */	0x20C0,
/*	MOVE.L	D1,(A0)	*/	0x2081
    };
   #pragma parameter __A0 pint64divu32(__A0,__D2)
    extern int64 *pint64divu32(int64*,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	BPL.S	*+6	;@1 */	0x6A04,
/*	NEG.L	(A0)	*/	0x4490,
/*	NEGX.L	D0	*/	0x4080,
/*@1	MOVE.L	(A0),D1 */	0x2210,
/*	MOVE.L	D2,(A0)	*/	0x2082,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A0),D2:D1 */	0x4C50, 0x1402,
/*	TST.L	-(A0)	*/	0x4AA0,
/*	BPL.S	*+6	;@2 */	0x6A04,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@2	MOVE.L	D0,(A0)	*/	0x2080,
/*	MOVE.L	D1,4(A0) */	0x2141, 4
    };
   #pragma parameter __D1 _uint64divu32(__A0,__D2)
    extern uint32 _uint64divu32(uint64*,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	MOVE.L	(A0),D1	*/	0x2210,
/*	MOVE.L	D2,(A0)	*/	0x2082,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A0),D2:D1 */	0x4C50, 0x1402,
/*	MOVE.L	D1,(A0)	*/	0x2081,
/*	MOVE.L	D0,-(A0) */	0x2100
    };
   #pragma parameter __A0 puint64divu32(__A0,__D2)
    extern uint64 *puint64divu32(uint64*,uint32) = {
	0x2018,
	0x2210,
	0x2082,
	0x4C42, 2,
	0x4C50, 0x1402,
	0x2081,
	0x2100
    };

   #pragma parameter __A0 _divi32(__A0,__D2)
    extern int32 _divi32(int64*,uint32) = {
/*	MOVE.L	D2,-(SP) */	0x2F02,
/*	BPL.S	*+8	;@1 */	0x6A06,
/*	NEG.L	D2	*/	0x4497,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@1	MOVE.L	4(A0),D1 */	0x2228, 4,
/*	MOVE.L	(A0),D0	*/	0x2010,
/*	BPL.S	*+6	;@2 */	0x6A04,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@2	TDIVU.L	D2,D2:D0 */	0x4C57, 2,
/*	DIVU.L	(SP)+,D2:D1 */	0x4C5F, 0x1402,
/*	TST.L	(A0)	*/	0x4A90,
/*	BPL.S	*+6	;end */	0x6A04,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080
    };
   #pragma parameter __D1 _divu32(__A0,__D2)
    extern uint32 _divu32(const uint64*,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	MOVE.L	(A0),D1	*/	0x2210,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A7)+,D2:D1 */	0x4C5F, 0x1402
    };

   /* returns low 32 bits */
   #pragma parameter __D2 _muls64s32(__A0,__D1,__D2)
    extern int32 _muls64s32(int64*,int32,int32) = {
/*	MULS.L	D1,D1:D2 */	0x4C01, 0x2C01,
/*	MOVE.L	D1,(A0)+ */	0x20C1,
/*	MOVE.L	D2,(A0)	*/	0x2082
    };
   #pragma parameter __D2 _mulu64s32(__A0,__D1,__D2)
    extern uint32 _mulu64s32(uint64*,int32,int32) = {
	0x4C01, 0x2C01,
	0x20C1,
	0x2082
    };
   #pragma parameter __D2 _muls64u32(__A0,__D1,__D2)
    extern int32 _muls64u32(int64*,uint32,uint32) = {
/*	MULU.L	D1,D1:D2 */	0x4C01, 0x2401,
/*	MOVE.L	D1,(A0)+ */	0x20C1,
/*	MOVE.L	D2,(A0)	*/	0x2082
    };
   #pragma parameter __D2 _mulu64u32(__A0,__D1,__D2)
    extern uint32 _mulu64u32(uint64*,uint32,uint32) = {
	0x4C01, 0x2401,
	0x20C1,
	0x2082
    };

   /* returns low 32 bits */
   #pragma parameter __D2 _int64muls32(__A0,__D2)
    extern int32 _int64muls32(int64*,int32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	TST.L	D2	*/	0x4A82,
/*	BPL.S	*+8	;@1 */	0x6A06,
/*	NEG.L	D2	*/	0x4482,
/*	NEG.L	(A0)	*/	0x4490,
/*	NEGX.L	D0	*/	0x4080,
/*@1	MULU.L	D2,D0	*/	0x4C02, 0,
/*	MULU.L	(A0),D1:D2 */	0x4C10, 0x2401,
/*	ADD.L	D0,D1	*/	0xD280,
/*	MOVE.L	D2,(A0)	*/	0x2082,
/*	MOVE.L	D1,-(A0) */	0x2101
    };
   #pragma parameter __D2 _uint64muls32(__A0,__D2)
    extern uint32 _uint64muls32(uint64*,int32) = {
	0x2018,
	0x4A82,
	0x6A06,
	0x4482,
	0x4490,
	0x4080,
	0x4C02, 0,
	0x4C10, 0x2401,
	0xD280,
	0x2082,
	0x2101
    };
   #pragma parameter __D2 _int64mulu32(__A0,__D2)
    extern int32 _int64mulu32(int64*,uint32) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	MULU.L	D2,D0	*/	0x4C02, 0,
/*	MULU.L	(A0),D1:D2 */	0x4C10, 0x2401,
/*	ADD.L	D0,D1	*/	0xD280,
/*	MOVE.L	D2,(A0)	*/	0x2082,
/*	MOVE.L	D1,-(A0) */	0x2101
    };
   #pragma parameter __D2 _uint64mulu32(__A0,__D2)
    extern uint32 _uint64mulu32(uint64*,uint32) = {
	0x2018,
	0x4C02, 0,
	0x4C10, 0x2401,
	0xD280,
	0x2082,
	0x2101
    };

   /* returns low 32 bits */
   #pragma parameter __D0 _int64neg(__A0)
    extern int32 _int64neg(int64*) = {
/*	NEG.L	4(A0)	*/	0x44A8, 4,
/*	MOVE.L	4(A0),D0 */	0x2028, 4,
/*	NEGX.L	(A0)	*/	0x4090
    };
   #pragma parameter __D0 _uint64neg(__A0)
    extern uint32 _uint64neg(uint64*) = {
	0x44A8, 4,
	0x2028, 4,
	0x4090
    };

   /* returns low 32 bits */
   #pragma parameter __D1 _int64half(__A0)
    extern int32 _int64half(int64*) = {
/*	MOVEQ	#0,D2 */	0x7400,
/*	MOVE.L	(A0),D0 */	0x2010,
/*	ADD.L	D0,(A0)+ */	0xD198,
/*	MOVE.L	(A0),D1 */	0x2210,
/*	ADDX.L	D2,D1 */	0xD382,
/*	ADDX.L	D2,D0 */	0xD182,
/*	ASR.L	#1,D0 */	0xE280,
/*	ROXR.L	#1,D1 */	0xE291,
/*	MOVE.L	D1,(A0) */	0x2081,
/*	MOVE.L	D0,-(A0) */	0x2100
   };
   #pragma parameter __D0 _uint64half(__A0)
    extern uint32 _uint64half(uint64*) = {
/*	MOVE.L	(A0)+,D0 */	0x2018,
/*	MOVE.L	(A0),D1 */	0x2210,
/*	LSR.L	#1,D0 */	0xE288,
/*	ROXR.L	#1,D1 */	0xE291,
/*	MOVE.L	D1,(A0) */	0x2081,
/*	MOVE.L	D0,-(A0) */	0x2100
   };

   #pragma parameter __D1 imuldiv(__D0,__D1,__D2)
    extern int32 imuldiv(int32,int32,int32) = {
/* 	MULS.L	D0,D0:D1 */	0x4C00, 0x1C00,
/* 	BPL.S	*+8	;@1 */	0x6A06,
/*	NEG.L	D2	*/	0x4482,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@1	MOVE.L	D2,-(A7) */	0x2F02,
/*	BPL.S	*+4	;@2 */	0x6A02,
/*	NEG.L	(A7)	*/	0x4497,
/*@2	TDIVU.L	(A7),D2:D0 */	0x4C57, 2,
/*	DIVU.L	(A7),D2:D1 */	0x4C57, 0x1402,
/*	TST.L	(A7)+	*/	0x4A9F,
/*	BPL.S	*+4	;end */	0x6A02,
/*	NEG.L	D1	*/	0x4481
    };
   #pragma parameter __D1 umuldiv(__D0,__D1,__D2)
    extern uint32 umuldiv(uint32,uint32,uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A7)+,D2:D1 */	0x4C5F, 0x1402
    };
  #endif
 #else /*little endian, no long long*/
  typedef struct int64{uint32 lo; int32 hi;} int64;
  typedef struct uint64{uint32 lo, hi;} uint64;
 #endif

 #define int64half(n)		(void)_int64half(&(n))
 #define uint64half(n)		(void)_uint64half(&(n))

 #define int64addhl(x,h,l)	_int64addhl(&(x),h,l)
 #define uint64addhl(x,h,l)	_uint64addhl(&(x),h,l)
 #define int64subhl(x,h,l)	_int64subhl(&(x),h,l)
 #define uint64subhl(x,h,l)	_uint64subhl(&(x),h,l)
 #define int64addhls(x,h,l)	_int64addhls(&(x),h,l)
 #define uint64addhls(x,h,l)	_uint64addhls(&(x),h,l)
 #define int64divu32(x,y)	_int64divu32(&(x),y)
 #define uint64divu32(x,y)	_uint64divu32(&(x),y)
 #define divi32(x,i)		_divi32(&(x),i)
 #define divu32(x,i)		_divu32(&(x),i)
 #define muls64s32(n,a,b)	_muls64s32(&(n),a,b)
 #define muls64u32(n,a,b)	_muls64u32(&(n),a,b)
 #define mulu64s32(n,a,b)	_mulu64s32(&(n),a,b)
 #define mulu64u32(n,a,b)	_mulu64u32(&(n),a,b)
 #define int64muls32(n,x)	_int64muls32(&(n),x)
 #define uint64mulu32(n,x)	_uint64mulu32(&(n),x)
 #define int64neg(n)		_int64neg(&(n))
 #define uint64neg(n)		_uint64neg(&(n))

 /* plain struct macros */
 #define int64hi(x)		(0+(x).hi)
 #define int64lo(x)		((uint32)(x).lo)

 #if BIG_ENDIAN
  #define _int64val(h,l)	{h,l}
  #define _int64xval(f)		{(f)/4294967296.0, \
	(f)-4294967296.0*((uint32)((f)/4294967296.0))}
 #else
  #define _int64val(h,l) {l,h}
  #define _int64xval(f)		{(f)-4294967296.0*((uint32)((f)/4294967296.0)), \
	(f)/4294967296.0}
 #endif

 #define _int64xset(x,f)	(void)((x).lo=(f)-4294967296.0*((uint32)((f)/ \
	4294967296.0)),(x).hi=(f)/4294967296.0)
 #define int64xset(x,f)		do{long double __f=f; \
	(x).lo=__f-4294967296.0*((x).hi=__f/4294967296.0);}while(0)
 #define int64sethl(x,h,l)	(void)((x).hi=h,(x).lo=l)
 #define int64setl(x,y)		do{long int _islv=y; \
	(x).hi=_islv>>32,(x).lo=_islv;}while(0)
 #define int64set(x,y)		(void)((x).hi=(y).hi,(x).lo=(y).lo)

 #define int64z(n)		(((n).hi|(n).lo)==0)
 #define int64n(n)		(((n).hi&0x80000000UL)!=0)

 #define int64toi(n)		((n).hi==(signed)(n).lo>>31? \
	(n).lo:0x7FFFFFFFL+((unsigned)(n).hi>>31))

 #define int64eq(x,y)		((x).hi==(y).hi&&(x).lo==(y).lo)
 #define int64eqhl(x,h,l)	(((x).hi==(h))&&((x).lo==(l)))
 #define int64gt(x,y)		((x).hi>(y).hi||((x).hi==(y).hi&&(x).lo>(y).lo))
 #define int64gthl(x,h,l)	((x).hi>(h)||((x).hi==(h)&&(x).lo>(l)))
 #define int64ge(x,y)		((x).hi>(y).hi||((x).hi==(y).hi&&(x).lo>=(y).lo))
 #define int64gehl(x,h,l)	((x).hi>(h)||((x).hi==(h)&&(x).lo>=(l)))
 #define uint64gtu32(x,i)	((x).hi!=0||(x).lo>(uint32)(i))

 #define int64add(x,y)		int64addhl(x,(y).hi,(y).lo)
 #define uint64add(x,y)		uint64addhl(x,(y).hi,(y).lo)
 #define int64addu32(x,i)	int64addhl(x,0,i)
 #define uint64addu32(x,i)	uint64addhl(x,0,i)

 #define int64sub(x,y)		int64subhl(x,(y).hi,(y).lo)
 #define uint64sub(x,y)		uint64subhl(x,(y).hi,(y).lo)
 #define int64subu32(x,i)	int64subhl(x,0,i)
 #define uint64subu32(x,i)	uint64subhl(x,0,i)

 #define int64divs32(x,i)	(((i)>=0)?int64divu32(x,i):_int64neg(pint64divu32(&(x),-(i))))
 #define uint64divs32(x,i)	(((i)>=0)?uint64divu32(x,i):_uint64neg(puint64divu32(&(x),-(i))))

 /* these only handle a literal count within [0,32] */
 #if BIG_ENDIAN
  #define _int64shr32(x,c)	((c)==16? \
	*(int32*)(2+(uint8*)&(x).hi):(((x).hi<<32-(c))|((x).lo)>>c))
  #define _uint64shr32(x,c)	((c)==16? \
	*(uint32*)(2+(uint8*)&(x).hi):(((x).hi<<32-(c))|((x).lo)>>c))
 #else
  #define _int64shr32(x,c)	(int32)(((x).hi<<32-(c))|((x).lo)>>c))
  #define _uint64shr32(x,c)	(uint32)(((x).hi<<32-(c))|((x).lo)>>c))
 #endif

 #define int64shl(n,i)		do{uint8 __i=(i)&0x7F; uint32 __t=(n).lo; \
	if((__i&=127)<=32) (n).lo=__t<<__i,(n).hi=((n).hi<<__i)|(__t>>32-__i); \
	else if((__i-=32)>32) (n).hi=0,(n).lo=0; else (n).hi=__t>>=__i,(n).lo=0; \
}while(0)
 #define int64shr(n,i)		do{uint8 __x=(i)&0x7F; uint32 __t=(n).hi; \
	if(__x<=31) (n).lo=(__t<<32-__x)|((n).lo>>__x),(n).hi=(signed)__t>>__x; \
	else if((__x-=32)<=31) (n).lo=(n).hi>>__x; else (n).hi=(n).lo=(n).hi>>32; \
}while(0)
 #define uint64shr(n,i)		do{uint8 __x=(i)&0x7F; uint32 __t=(n).hi; \
	if(__x<=31) (n).lo=(__t<<32-__x)|((n).lo>>__x),(n).hi=__t>>__x; \
	else if((__x-=32)<=31) (n).lo=(n).hi>>__x; else (n).hi=(n).lo=(n).hi>>32; \
}while(0)

 #ifndef int64half
  #define int64half(n)		do{int64 *__n=&(n); \
	if((__n->hi&(long)0x80000000)!=0&&++__n->lo==0) ++__n->hi; \
	__n->lo=(__n->lo>>1)+(__n->hi<<31),__n->hi>>=1; \
}while(0)
  #define uint64half(n)		do{uint64 *__n=&(n); \
   __n->lo=(__n->lo>>1)+(__n->hi<<31),__n->hi>>=1; \
}while(0)
 #endif

/* -- commented out because of the conflicting nature of parameter usage
 #ifndef int64addhl
  #define int64addhl(x,h,l)	do{int64 *__x=&(x); uint32 __l=__x->lo; \
	__x->hi+=h; if((__x->lo+=l)<__l) ++__x->hi; \
}while(0)
  #define uint64addhl(x,h,l)	do{uint64 *__x=&(x); uint32 __l=__x->lo; \
	__x->hi+=h; if((__x->lo+=l)<__l) ++__x->hi; \
}while(0)
 #endif
*/

 #define int64or(x,y)		((x).hi|=(y).hi,(x).lo|=(y).lo)
 #define int64eor(x,y)		((x).hi^=(y).hi,(x).lo^=(y).lo)
 #define int64and(x,y)		((x).hi&=(y).hi,(x).lo&=(y).lo)
 #define int64not(x)		((x).hi=~(x).hi,(x).lo=~(x).lo)

 #define int64orhl(x,h,l)	(x,(x).hi|=h,(x).lo|=l)
 #define int64eorhl(x,h,l)	(x,(x).hi^=h,(x).lo^=l)
 #define int64andhl(x,h,l)	(x,(x).hi&=h,(x).lo&=l)
#endif /*!LONG_LONG*/

 /* universal 64 bit macro... params used twice */
#define int64tou(n)		(-(int64hi(n)!=0)|int64lo(n))

#define __ENVIRO64_H
#endif
