/*
 macos/menus.c (formerly macmenus.c)

  Created: 2008.07.08
  Revised: 2010.04.11

Wednesday, August 19, 2009 1:10 AM
  increased/maxed out some option limits (1000-65535 srate, 24-429% for speed).
Friday, October 9, 2009 1:48 PM
  finishing the menubar show/hide code to work properly.
Tuseday, March 23, 2010 1:26 PM
  moved all the rest of the sound stuff declared here out, externs are now in shell.h.
Saturday, March 27, 2010 1:51 PM
  a run-in with a falsely rewritten "Set Enabled" routine prompts me to check validity.
*/

#include <stdlib.h>

#include "shell.h"

#include "../FCEU.h"
#include "../common/dface.h"

#include "shellmenus.h"
#include "messagelog.h"
#include "macdialogs.h"
#include "macvideo.h"
#include "debugger.h"
#include "config.h"
#include "tileview.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- globals */

extern struct FCEUGI *GI;

static const uint16	sndqualities[] =
{
	8192,
	11025,
	15000,
	22050,
	36000,
	48000,
	64000,
/*	96000+(0<<24),
	44100+(1<<24),
	48000+(1<<24),
	96000+(1<<24),
	44100+(2<<24),
	48000+(2<<24),
	96000+(2<<24),*/
};

static const uint32	emulationspeeds[] =
{
	500000000,
	629960525,
	707106781,
	793700526,
	890898718,
	1000000000,
	1259921050,
	1587401052,
	1781797436,
	2000000000,
};

/* ---- code */

void redomenusgi(FCEUGI *gi)
{ /* close, load, save, reset, power, step (,cheat) */
	uint8		*simpled = (uint8*)"\102\105\104\201\202\205\0";

	int		x = -1;
	uint8		i;

	while((i=simpled[++x])!=0)
		enablemenuitem(i,gi!=NULL);
}

void reenablemenus(boolean enable)
{
	uint8		*simpled = (uint8*)"\1\4\2\3\0";

	int		x = -1;
	uint8		i;

	while((i=simpled[++x])!=0)
		enablemenuitem(i<<6,enable);
	enablemenuitem(0001,enable); /* was the whole menu; 2010.04.11 */

	drawmenubar();
}

void updatemenus()
{
	static uint8	*polo[] = {&ConsolePitchFix,&genie,&fullscreen,
		&screenshift,&screenrefreshsync,&fceutileview};
	uint16		simpled[] = {0307, 0311, 0701, 0702, 0706, 0710, 0};

	int		x = -1, m;
	uint16		i;

	while((i=simpled[++x])!=0)
		checkmenuitem(i,*polo[x]&1);

	//partial switch added to tell whether it's applicable; 2010.03.23, 03.24
	m = palyo&1;
	if (GI!=NULL && GI->vidsys!=GIV_USER
			&&((GI->vidsys!=GIV_NTSC && m==0)
			|| (GI->vidsys!=GIV_PAL && m!=0)))
		m|= 0x8000U;
	checkmenuitem(0310,m);

	for(x=1; x<=sizeof(sndqualities)/sizeof(sndqualities[0]); ++x)
		if (sndqualities[x-1]==soundrate)
			break;
	checkmenu(0501,x,0); /* if loop fails, Other will automatically be checked */

	if ((x=soundvolume/10)*10==soundvolume && (unsigned)x<=15)
		checkmenu(0601,++x,0);
	else
		checkmenu(0601,17,0);

	checkmenuitem(0704,(eoptions&EO_CLIPSIDES)!=0);
	checkmenuitem(0705,(eoptions&EO_NOSPRLIM)!=0);

	for(x=01001,m=1; x<=01005; ++x,m+=m)
		checkmenuitem(x,sndactivechannels&m);

	for(x=1; x<=sizeof(emulationspeeds)/sizeof(emulationspeeds[0]); ++x)
		if (emulationspeeds[x-1]==EmulationSpeedRatio)
			break;
	checkmenu(01101,x,0);

	checkmenu(01201,1+(eoptions&EO_BGRUN)+((eoptions&EO_MUTE) ? 2:0),0);
}

boolean domenu(uint16 itemcode)
{
	int		tmp;

	switch(itemcode)
	{ /* menubar toggle is automatic during fullscreen */
	/* apple */
	case 0001: /* about */
		stopsound();
		splash();
		break;

	/* file */
	case 0101: /* open */
		stopsound();
		opengame(NULL);
		break;
	case 0102: /* close */
		scexit(_sc_closegame);
		break;
	case 0104: /* save */
		if (GI==NULL)
			return false;
		stopsound();
		savestate(NULL);
		break;
	case 0105: /* load */
		if (GI==NULL)
			return false;
		stopsound();
		loadstate(NULL);
		break;
		/* - */
	case 0107: /* record */
		setmenuitemstr(0107,recordsoundwave(NULL) ?
			"\pStop Sound Logging":"\pLog Sound As...");
		break;
		/* - */
	case 0111: /* quit */
		scexit(_sc_quit);
		break;

	/* emulator */
	case 0201: /* reset */
		FCEUI_ResetNES();
		goto subreset;
	case 0202: /* power */
		FCEUI_PowerNES();
	subreset:
		if (debuggerwindow_visible())
			debuggerwindow_step(1); /* start */
		break;
	case 0204: /* debug */
		checkmenuitem(0204,debuggerwindow_show(-1)>0);
		break;
	case 0205: /* step */
		checkmenuitem(0204,debuggerwindow_step(1)>0);
		break;

	/* game */
	/*case 0301: -- sound submenu */
	/*case 0302: -- volume submenu */
	/*case 0303: -- channels submenu */
	/*case 0304: -- video submenu */
	/*case 0305: -- speed submenu */
		/* - */
	case 0307: /* PAL rate fix */
		ConsolePitchFix^= 1;
		RefreshThrottleFPS();
		InitSound(GI);
		break;
	case 0310: /* palyo */
		stopsound();
		FCEUI_SetVidSystem(palyo^=1);
		RefreshThrottleFPS(); //first to clear
		videorefresh();
		RefreshThrottleFPS(); //second to ensure timing update
		InitSound(GI);
		break;
	case 0311: /* genie */
		FCEUI_SetGameGenie(genie^=1);
		break;
	/*case 0312: -- background submenu */
		/* - */
	case 0314: /* input */
		stopsound();
		config_inputdialog();
		break;
	case 0315: /* palette */
		stopsound();
		gamewindowtofront();
		PaletteDialog(NULL,0);
		break;
	case 0316: /* netplay */
		stopsound();
/*		ShowNetplayConsole();*/
		break;
	/* cheat item replaced with commandline commands (will be implemented eventually) */

	/* edit */
	case 0410: /* log */
		checkmenuitem(0410,logwindow_show(-1)>0);
		break;

	/* sound quality submenu */
	case 0501: /* lowest */
	case 0502: /* AM Radio */
	case 0503: /* low */
	case 0504: /* default */
	case 0505: /* moderate */
	case 0506: /* good */
	case 0507: /* better */
	case 0510: /* Other... */
	case 0511: /* CD-1 */
	case 0512: /* high-1 */
	case 0513: /* best-1 */
	case 0514: /* CD-2 */
	case 0515: /* high-2 */
	case 0516: /* best-2 */
		stopsound();

		if ((itemcode+=-0501)>=sizeof(sndqualities)/sizeof(sndqualities[0]))
		{
			if ((tmp=valuedialog("\pCustom SampleRate",
					soundrate,1000,96000,false))==INT_NO_VALUE)
				goto end;
			if (tmp>65535)
				if (showdialog(dialogid_confirm,confirm_yes,"\pConfirm SampleRate",
"Most Wave players dont recognize rates above 65535.  Set anyway?",
						NULL,NULL,NULL)==confirm_no)
				goto end;
			soundrate = tmp;
		} else
			soundrate = sndqualities[itemcode]&0xFFFF;
		if (GI!=NULL)
			GI->soundrate = 0;
		InitSound(GI);
		FCEUI_SetSoundQuality(soundquality=0); /*(qualities[itemcode]>>24)&7);*/
		break;

	/* sound volume submenu (sorry, no slider) */
	case 0601: /* off */
	case 0602: /* 10% */
	case 0603: /* 20% */
	case 0604: /* 30% */
	case 0605: /* 40% */
	case 0606: /* 50% */
	case 0607: /* 60% */
	case 0610: /* 70% */
	case 0611: /* 80% */
	case 0612: /* 90% */
	case 0613: /* 100% */
	case 0614: /* 110% */
	case 0615: /* 120% */
	case 0616: /* 130% */
	case 0617: /* 140% */
	case 0620: /* 150% */
	case 0621: /* Other... */
		if ((itemcode+=-0601)==0)
		{
		zerovolume:
			soundvolume = 0; /* now must set ahead of InitSound */
			stopsound();
			InitSound(NULL);
		}

		if (itemcode>15)
		{ /* negative sound volumes can still be accomplished in the Config file */
			if ((tmp=valuedialog("\pCustom Sound Volume",
					soundvolume,0,10000,false))==INT_NO_VALUE)
				goto end;
			if ((soundvolume=tmp)==0)
				goto zerovolume;
		} else
			soundvolume = itemcode*10;

		if (soundon && soundvolume>0)
			InitSound(GI);

		FCEUI_SetSoundVolume(soundvolume);
		break;

	/* video */
	case 0701: /* fullscreen */
		if (GI==NULL)
			fullscreen^= 1;
		else
			initializescreen(fullscreen^1);
		checkmenuitem(0701,fullscreen);
		goto end;
	case 0702: /* 2x scale */
		checkmenuitem(0702,screenshift^=1);
		if (GI!=NULL)
			initializescreen(fullscreen);
		videorefresh();
		goto end;
		/* - */
	case 0704: /* clip */
		eoptions^= EO_CLIPSIDES;
		videorefresh();
		break;
	case 0705: /* sprite unlimit */
		eoptions^= EO_NOSPRLIM;
		videorefresh();
		break;
	case 0706: /* refresh sync */
		screenrefreshsync^= 1;
		break;
	case 0710: /* tile view */
		ShowHideTileView(-1);
		break;

	/* sound channels */
	case 01001:
	case 01002:
	case 01003:
	case 01004:
	case 01005:
		sndactivechannels^= 1<<itemcode-01001;
		FCEUI_SetSoundActiveChannels(sndactivechannels);
		break;
		/* - */
	case 01007: /* Square/Rect Amp */
		if ((tmp=valuedialog("\pRectangle Filter Amplitude (256 = 1.0)",
				rectangle_amp,-10000,10000,false))!=INT_NO_VALUE)
			FCEUI_SetRectFilterAmp(rectangle_amp=tmp);
		goto end;

	/* speed */
	case 01101: /* 50% */
	case 01102: /* 63% */
	case 01103: /* 71% */
	case 01104: /* 79% */
	case 01105: /* 89% */
	case 01106: /* 100% */
	case 01107: /* 126% */
	case 01110: /* 159% */
	case 01111: /* 178% */
	case 01112: /* 200% */
	case 01113: /* Other... */
		if ((itemcode+=-01101)>=sizeof(emulationspeeds)/sizeof(emulationspeeds[0]))
		{
			double		td;

			if ((td=valuedialog("\pCustom Emulation Speed",
					EmulationSpeedRatio/10000000.0,24,429,true))==INT_NO_VALUE)
				goto end;
			EmulationSpeedRatio = td*10000000.0+.5;
		} else
			EmulationSpeedRatio = emulationspeeds[itemcode];
		RefreshThrottleFPS();
		InitSound(GI);
		break;

	/* background */
	case 01201: /* off */
	case 01202: /* full */
	case 01203: /* muted */
	case 01204: /* bg muted */
		eoptions&=~(EO_BGRUN|EO_MUTE);
		if (--itemcode&1)
			eoptions|= EO_BGRUN;
		if (itemcode&2)
			eoptions|= EO_MUTE;
	}

	updatemenus();
end:
	return true;
}
