/*
 macos/shell.c
  main code for MacFCEU
*/
#define REVISION_DATE			"2010.04.11 15:20 EDT"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <Files.h>
#include <Dialogs.h>
#include <Gestalt.h>
#include <AppleEvents.h>
#include <Sound.h>

#include "shell.h"

#include "../FCEU.h"
#include "../common/dface.h"
#include "../PPU.h"

#include "shellmenus.h"
#include "macdialogs.h"
#include "messagelog.h"
#include "macvideo.h"
#include "debugger.h"
#include "config.h"
#include "tileview.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- constants */

#ifndef NO_CONFIG
 #define CONFIGFILE_PATH		"macfceu.cfg"
#endif

/* ---- globals */

WindowRef	myfrontwindow;

char		temporary[1024] = {0};

static sc_type	softclose = 0;
boolean		suspended = false, mousehidden = false;
boolean		soundon = 1, soundavailable;

short		vmenubar = 20;

static char	defaultdirs[][256] =
{
	":cheats",
	":gameinfo",
	":sav",
	":fcs",
	":snaps",
	""	//was ":", couldn't really work as result; 2010.03.27
};

uint8		cpalette[192];

FCEUGI		*GI = NULL;
int		eoptions = EO_SNAPNAME; /* EO_SNAPNAME now on by default */
int		soundrate = 22050;
int		soundvolume = 100;
int		soundquality = 0;

int		ntsccol, ntsctint, ntschue;

uint8		fullscreen = 1, screenshift = 0, genie = 0, palyo = 0;
boolean		screenrefreshsync;

uint8		sndactivechannels = 0x1F;
int		rectangle_amp = 256;

uint8		*xbsave = NULL; /* for manual updates */

/* timing */

static int	maxconbskip = 9; /* maximum consecutive blitter skips */
static int	ffbskip = 9; /* blit skips per blit when fast-forwarding */

/* ---- messages */

void FCEUD_Message(const char *s)
{
	logmessage(s,strlen(s));
}

void FCEUD_PrintError(const char *s)
{
	logmessage(s,strlen(s));
	logmessage("\n",1);
	if (mousehidden)
		{InitCursor(); mousehidden = false;}
	showdialog(dialogid_error,std_ok,"\pFCE Ultra Error",s,"\p",NULL,NULL);
	/* mouse cursor will hide automatically, if need be */
}

void splash()
{
	char		buf[200];

	sprintf(buf,
"MacFCEU " VERSION_STR "\n"
"Revised:   " REVISION_DATE "\n" /* __DATE__/__TIME__ are inconsistent (e.g., compiler date) */
"Compiler: " CVERSIONSTR "\n"
"http://macfceu.sourceforge.net\n"
"Initially based on FCE Ultra 0.98.13.\n"
"http://fceultra.sourceforge.net\n"
#ifdef __NVERS
		,__NVERS
#endif
		);
	showdialog(dialogid_about,1,"\pAbout MacFCEU",buf,"\p",NULL,NULL);
}

sc_type scexit(sc_type what)
{
	if (softclose==_sc_quit /* already terminating */
			|| what==0) /* since softclose is no longer public */
		goto end;
	if (what&0x80 && what!=_sc_quit)
		{softclose&= what; goto end;}

	if ((what==_sc_quit || what==(_sc_reopengame|_sc_closegame)) && GI!=NULL)
	{ /* only if a game is active */
		int		strc, stri;

		stopsound();

		stri = what==_sc_quit ? strlistid_exit:strlistid_close;
		if ((strc=countstrlist(stri))==0)
			temporary[512] = 0;
		else
			getcstr(512+temporary,stri,rand()%strc);
		/* if no str (or its empty), use default string(s) */
		if (showdialog(dialogid_confirm,confirm_no,
				temporary[512]==0 ? NULL:
					what!=_sc_quit ? "\pClose game?":"\pQuit during game?",
				temporary[512]==0 ? "Are you sure you want to exit?":
					512+temporary,
				"\p",NULL,NULL)!=confirm_yes)
			goto end;
	}
	if (what!=_sc_quit)
		what&= _sc_reopengame|_sc_closegame;
	softclose|= what;
end:
	return softclose;
}

static void printinfo(void)
{
	FCEUD_Message(
"  MacFCEU " VERSION_STR //"  -- WARNING: DEBUG build!" --time to retire this
#ifdef env68k
"\n   -- for 680x0 Macs, performance may be unacceptably slow"
#endif
			);
	if (!soundavailable)
		FCEUD_Message("\n  -- Sound Manager 3.0+ not present (no sound playback)");
	FCEUD_Message(
"\n\n"
"Key assignments (see menubar for other stuff):\n"
"   `    Fast forward.\n"
"  H/V   Select NTSC hue or tint (-/= to adjust).\n"
"  0-9   Show movie slot.\n"
"  F4    Hide sprite layer (shift=bkgnd).\n"
"  F5    Save state (shift=movie).\n"
"  F6    Show/Hide dips/disks.\n"
"  F7    Load state (shift=movie).\n"
"  F8    Insert/Eject coin | enter barcode.\n"
"  F9    Take snapshot.\n"
"  F10   Toggle FPS.\n"
/*"Gamepad(s):\n" -- now the input dialog tells you whats what
"   A/ B/Sl/St: Ctl Opt Tab  Q  |  /  Shf  '  Ret\n"
"  Up/Dn/Lf/Rg:  W   S   A   D  | AUp ADn ALf ARg\n"*/
"\n"
			);
}

/* ---- private game code */

static void closegame()
{
	debuggerwindow_clear(false);
	cleargamepath();
	if (GI==NULL)
		return;
/*	debuggerwindow_show(0);*/
	FCEUI_CloseGame();
	GI = NULL;
	redomenusgi(GI);
	debuggerwindow_clear(true);
}

/* ---- general code */

static void setdirectories(void)
{
	static int	jlist[6] =
	{
		FCEUIOD_CHEATS,FCEUIOD_MISC,FCEUIOD_NV,FCEUIOD_STATE,FCEUIOD_SNAPS,
		INT_NO_VALUE
	};
	char		*n;
	int		i;
	short		x = namesstr_cheatsdir;

	FCEUI_SetSnapName(eoptions&EO_SNAPNAME);

	do
	{
		n = defaultdirs[x-namesstr_cheatsdir];

		getcstr(n,strlistid_names,x);
		if ((i=jlist[x-1])==INT_NO_VALUE)
			break;
		mkdir(n); /* create directory */
		FCEUI_SetDirOverride(i,n);
	} while(++x!=0);

	if (x==0)
		return;
	FCEUI_SetBaseDirectory(n);
}

int setcursor(int ID, boolean color)
{
	if (ID<0)
		{InitCursor(); return 0;}
	if (color)
	{
		CCrsrHandle	crsr;

		if ((crsr=GetCCursor(ID))==NULL)
			return -1;
		SetCCursor(crsr);
		DisposeCCursor(crsr);
	} else
	{
		CursHandle	cursor;

		if ((cursor=GetCursor(ID))==NULL)
			return -1;
		SetCursor(*cursor);
	}

	return 1;
}

/* previously in menus.c */
int gamewindowtofront()
{
	if (gamewindow==NULL)
		return -1;
	if (fullscreen) /* in window mode only; 2009.10.06 22:02 EDT */
		return 0;

	BringToFront(gamewindow);

	return 1;
}

/* ---- shell menus */

/* used to be public */
static boolean reinitmenus(void)
{ /* apple, file, edit, emulator, special, [hierachical] */
	uint8		*simpled = (uint8*)"\0\1\4\2\3\205\206\207\210\211\212\377";

	int		x = -1;
	uint8		i;
	boolean		gotall = true;

	ClearMenuBar();
	while((i=simpled[++x])!=0xFF)
	{
		MenuHandle	menu;

		if ((menu=MacGetMenu(menuid_apple+(i&0x7F)))==NULL)
			{gotall = false; continue;}
		MacInsertMenu(menu,((i&0x80)!=0)*kInsertHierarchicalMenu);

		if (i==0)
			AppendResMenu(menu,'DRVR');
	}

	drawmenubar();

	return gotall;
}

static boolean handlemenu(long itemhit)
{
	HiliteMenu(0);

	/* sometimes the mouse remains hidden for dumb reasons; 2009.01.29 */
	mousehidden = true;

	if ((itemhit+=-menuid_apple<<16)&0xFC00FFC0)
		return false;
	if ((unsigned)(itemhit-2)<=32765)
	{
#if TARGET_API_MAC_CARBON
		return false; /* shouldnt come here */
#else
		GetMenuItemText(GetMenuHandle(menuid_apple),itemhit,(UInt8*)temporary);
		OpenDeskAcc((UInt8*)temporary);
		return true;
#endif
	}

	return domenu((itemhit>>10&~63)|(itemhit&63));
}

/* ---- keys */

static int handlekey(EventRecord *event, DialogRef dialog)
{
extern uint8 show_fps;

	long		menuitem;
	char		charcode = event->message&charCodeMask;

	if (event->modifiers&cmdKey)
	{
		if (dialog==NULL && (menuitem=MenuKey(charcode))!=0)
			{handlemenu(menuitem); return -1;}
		/* special (unreliable) cmd-keys: home, top of list; end, end of list */
		if (charcode==kHomeCharCode)
			return 0x7F20;
		else if (charcode==kEndCharCode)
			return 0x7FFF;
	}

	if (dialog!=NULL) /* major fix */
		return 0;

	if (charcode==kFunctionKeyCharCode)
	{
		switch(event->message&0xFF00)
		{
		case 0x7A00: /* F1 */
			domenu(0710);
			return -1;
		case 0x6D00: /* F10 */
			show_fps = !show_fps;
			return -1;
		case 0x7100: /* F15 */
			if (IsMenuItemEnabled(GetMenuHandle(menuid_emulator),0204))
				{checkmenuitem(0204,debuggerwindow_show(-1)>0); return -1;}
			break;
		}
	} else if (charcode==kEnterCharCode)
	{ /* note that enter is ignored when the debugger window is hidden/uninitialized */
		if (getvisiblewindowport(debuggerwindow)!=NULL)
		{ /* command-enter: step 16 */
			checkmenuitem(0204,debuggerwindow_step(event->modifiers&cmdKey ? 16:1)>0);
			return -1;
		}
	} else if ((charcode==kPageUpCharCode
			|| charcode==kPageDownCharCode
			|| charcode==kUpArrowCharCode
			|| charcode==kDownArrowCharCode)
			&& (gamewindow==NULL || getvisiblewindowport(debuggerwindow)!=NULL))
	{
		int		cmdcode, vcnt = charcode;

		if ((cmdcode=(event->modifiers&cmdKey)<<16)!=0
				&& charcode==kUpArrowCharCode)
			cmdcode = 0, vcnt = -1L<<16; /* hack scroll fix; 2010.04.10 */
		else
			vcnt|= cmdcode, vcnt|= 0x10000;
		logwindow_scroll(cmdcode>>8,vcnt);
		return -1;
	}

	return 0; /* invoke standard proc */
}

/* ---- events */

static void suspend(void)
{
	suspended = true;
	ShowMenuBar();
}

static void resume(void)
{
	suspended = mousehidden = false;
	mycolorsinvalidate = 2; /* update all colors upon return */
}

int handleevent(EventRecord *event, DialogRef dialog)
{
	Rect		bounds;
	WindowRef	window;
	int		hit, part;

	/* cursor show/hide moved here so menubar shows for all callers; 2009.10.08 08:55 EDT */
	if (!fullscreen)
	{
		if (mousehidden)
			{mousehidden = false; goto show;}
	} else if (gamewindow!=NULL)
		if (dialog!=NULL || event->where.v<vmenubar)
		{
		show:
			ShowMenuBar();
			if (mousehidden)
				{InitCursor(); mousehidden = false;}
		} else if (!suspended)
			{HideMenuBar(); mousehidden = false;} /* to hide cursor */

	myfrontwindow = suspended ? NULL:FrontWindow();

	if (event==NULL)
		return 0;

	hit = -1;
	switch(event->what)
	{
	case mouseDown:
		switch(part=FindWindow(event->where,&window))
		{
		/*case inNoWindow:
			return -1;*/
		case inMenuBar:
			handlemenu(MenuSelect(event->where));
			break;
		case inDrag:
			if (dialog!=NULL && GetDialogWindow(dialog)!=window
					&& (event->modifiers&cmdKey)==0)
				{SysBeep(20); break;}
			/*if (fullscreen)
				break;*/
			/* the Mac OS writers made a mistake in taking the keys without option
			  in DragWindow; this is the very reason palettes need extra shim code */
			/* heres somethin: letting go of the commandkey at the right time with
			  a dialog in front can give you the ability to front other windows */
			GetRegionBounds(GetDragGrayBounds(),&bounds);
			DragWindow(window,event->where,&bounds);
			hit = -3;
			break;
		case inGrow:
		case inContent:
			if (dialog!=NULL && GetDialogWindow(dialog)!=window)
				{SysBeep(20); break;}
			if (!fullscreen)
				SelectWindow(window);
			if (window==logwindow)
				{logwindow_click(event,part); break;}
			else if (dialog!=NULL && GetDialogWindow(dialog)==window
					&& (hit=dialogselect(dialog,event))==0)
				hit = -1; /* dont call standardmodal twice */
			else if (window==tilewindow)
				TileViewClick(event);
			else if (window==debuggerwindow)
				if (part==inGrow)
					debuggerwindow_grow(event);
				else
					debuggerwindow_click(event);
			break;
		case inGoAway:
			if (dialog!=NULL && GetDialogWindow(dialog)!=window)
				{SysBeep(20); break;}
			if (!TrackGoAway(window,event->where))
				break;
			if (window==gamewindow)
				softclose|= _sc_closegame;
			else if (window==logwindow)
				{checkmenuitem(0410,logwindow_show(0)>0); break;}
			else if (window==debuggerwindow)
				{checkmenuitem(0204,debuggerwindow_show(0)>0); break;}
			else if (window==tilewindow)
				domenu(0710);
			else
				HideWindow(window);
			hit = -2; /* signal close */
		}
		break;
	case keyDown:
	case autoKey:
		hit = handlekey(event,dialog);
		break;
	case activateEvt:
		HiliteMenu(0);

		/* one thing to piss people off; some dialogs that cut in front may
		  modal-disable the buttons on the program dialog */
		hit = 0;

		window = (WindowRef)event->message;
		if (dialog==NULL)
			if (window==logwindow)
				logwindow_activate(event->modifiers&activeFlag);
			else if (window==debuggerwindow)
				debuggerwindow_activate(event->modifiers&activeFlag);
		break;
	case updateEvt:
		if ((window=(WindowRef)event->message)==NULL)
			break;
		BeginUpdate(window);
		if (window==gamewindow)
			FCEUD_BlitScreen(xbsave); /* update whatever we can */
		else if (window==logwindow)
			logwindow_update(true);
		else if (window==debuggerwindow)
			debuggerwindow_update(true);
		else if (window==tilewindow)
			BlitTileWindow(1);
		else if (GetDialogWindow(dialog)==window)
			dialogupdate(dialog);
		EndUpdate(window);
		break;
	case osEvt:
		if ((event->message&(long)0xFF000000)!=suspendResumeMessage<<24)
			break;
		InitCursor();

		if (event->message&resumeFlag)
			resume();
		else
			suspend();
		break;
	case kHighLevelEvent:
		AEProcessAppleEvent(event);
		break;

	default:
		hit = 0;
	}

	return hit;
}

#include "config.c"

/* ---- Apple Events[TM] */

boolean		handlersinstalled = false;

static OSErr eventcheck(const AppleEvent *event)
{
	DescType	type;
	Size		actual;
	OSErr		err;

	if ((err=AEGetAttributePtr(event,keyMissedKeywordAttr,typeWildCard,
			&type,0,0,&actual))==errAEDescNotFound)
		return noErr;

	return err==noErr ? errAEEventNotHandled:err;
}

static pascal OSErr odocevent(const AppleEvent *event,
	AppleEvent *reply, 
	long refcon)
{
#pragma unused(reply,refcon)
	AEDescList	list;
	long		index, count;
	OSErr		err;
	boolean		first = true;

	if ((err=AEGetParamDesc(event,keyDirectObject,typeAEList,&list))!=noErr
			|| (err=eventcheck(event))!=noErr
			|| (err=AECountItems(&list,&count))!=noErr)
		return err;

	for(index=1; index<=count; ++index) 
	{
		FSSpec		file;
		Size		actual;
		DescType	type;
		AEKeyword	keyword;

		if ((err=AEGetNthPtr(&list,index,typeFSS,&keyword,
				&type,(Ptr)&file,sizeof(FSSpec),&actual))!=noErr)
			continue;
		if (takefile(&file)==0 && first)
			scexit(/*GetKeyboard()[6]&0x80 ?
					_sc_reopengame:*/_sc_reopengame|_sc_closegame),
			first = false;
	}

	return noErr;
}

static pascal OSErr oappevent(const AppleEvent *event,
	AppleEvent *reply, 
	long refcon)
{
#pragma unused(reply,refcon)
	return eventcheck(event);
}

static pascal OSErr quitevent(const AppleEvent *event,
	AppleEvent *reply, 
	long refcon)
{
#pragma unused(reply,refcon)
	OSErr		err;

	if ((err=eventcheck(event))!=noErr)
		return err;
	softclose = _sc_quit;

	return noErr;
}

static pascal OSErr pdocevent(const AppleEvent *event,
	AppleEvent *reply, 
	long refcon)
{
#pragma unused(reply,refcon)
	OSErr		err;

	if ((err=eventcheck(event))!=noErr)
		return err;
	return errAEEventNotHandled;
}

static void installhandlers(void)
{
	AEEventHandlerUPP	handler;

	if (handlersinstalled)
		return;

	/* ordered in importance in case any fail */
	if ((handler=NewAEEventHandlerUPP(odocevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEOpenDocuments,handler,0,false);
	if ((handler=NewAEEventHandlerUPP(oappevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEOpenApplication,handler,0,false);
	if ((handler=NewAEEventHandlerUPP(pdocevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEPrintDocuments,handler,0,false);
	if ((handler=NewAEEventHandlerUPP(quitevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEQuitApplication,handler,0,false);

	handlersinstalled = true;
}

static void tryevents(void)
{
	EventRecord	event;
	int		retries = 60;

	do
		if (WaitNextEvent(highLevelEventMask,&event,0,NULL))
			AEProcessAppleEvent(&event);
	while(--retries!=0);
}

/* ---- system check */

static void prerequisites(long *type)
{
	UInt8		buf[64];

/* screw checks for whether Gestalt is implemented */
	*type = 0;
	if (Gestalt(gestaltSystemVersion,type)!=noErr || *type<0x710)
	{ /* since we might be starting on a 68000, sprintf should be avoided here (32 bit ins.) */
		UInt8		*d = buf;
		unsigned	value = *type;

		*d++ = 12;
		*d++ = '(';
		*d++ = 'u'; *d++ = 's'; *d++ = 'i'; *d++ = 'n'; *d++ = 'g'; *d++ = ' ';
		if ((*d=value>>12&15)!=0)
			*d+++= '0', *d = value>>8&15, ++buf[0];
		*d+++= '0';
		*d++ = '.';
		*d++ = '0'+((value>>4)&15);
		*d++ = '0'+(value&15);
		*d = ')';
		getpstr(temporary,strlistid_errors,errorstr_oldsystem);
		goto dlg;
	}
	*type = -1; /* selector should exist under 7.1+ */
	if (Gestalt(gestaltNativeCPUtype,type)!=noErr || *type<gestaltCPU68040)
	{
		if (*type<0)
			buf[0] = 0;
		else
			buf[0] = sprintf(1+(char*)buf,"(using MC680%c0)", *type+'0');
		getpstr(temporary,strlistid_errors,errorstr_oldcpu);
	dlg:
		showalert(alertid_fatal,(UInt8*)temporary,buf,"\p");
	}
}

#if !TARGET_API_MAC_CARBON
extern void installPatches(void);
#endif

static void initialize(long int stackhigh)
{
#pragma unused(argc,argv)
	long		value = 0;

	MoreMasters();
	MoreMasters();
	MoreMasters();

# if TARGET_API_MAC_CARBON /* Carbon should already add more than enough stack space */
#  pragma unused(stackhigh)
# else
	SetApplLimit((Ptr)((-64*1024)+stackhigh)); /* adds about 64KB of stack space */
	MaxApplZone();
	InitGraf((Ptr)&qd.thePort);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0);
# endif
	prerequisites(&value);

//	remove("stderr");

	InitCursor();

	/* install AE handlers for dragging in files */
	if (Gestalt(gestaltAppleEventsAttr,&value)==noErr
			&& (value&(1<<gestaltAppleEventsPresent)))
		installhandlers();

	gamegdevice = GetMainDevice();

	srand(TickCount());
#if !TARGET_API_MAC_CARBON
	installPatches();
#endif
}

#ifndef TERM_CLRALL
 #define TERM_CLRALL 1
#endif
#if TERM_CLRALL
 #include "usercommands.h"
#endif

static void cleanup(void)
{
#ifdef CONFIGFILE_PATH
	configfile_save(CONFIGFILE_PATH);
#endif
	closegame();
	FCEUI_Kill();
	KillSound();
	FCEUI_EndWaveRecord();
#if TERM_CLRALL
	commandline_free();
#endif
	InitCursor();
	FlushEvents(mDownMask|mUpMask|keyDownMask|keyUpMask|autoKeyMask,0);
}

/* ---- main */

void FCEUD_Update(uint8 *XBuf,
	int16 *Buffer,
	int Count);

static int eventloop(void)
{
	int		result = 0;

	if (getnextgamepath()==NULL) /* fullscreen by default when loading by HLE (Finder) */
		fullscreen = 0;

	updatemenus();
	redomenusgi(NULL);
	enablemenuitem(0706,VideoRetraceSync((GDHandle)-1)!=paramErr);

	printinfo(); /* print info regardless (now); 2009.01.18 */
	if (getnextgamepath()==NULL)
	{
		domenu(0410); /* show info (message log) if not in fullscreen */
		if (GetKeyboard()[6]&0x80) /* command key; was (eoptions&EO_FOAFTERSTART) */
			opengame(NULL);
	}

	while(softclose!=_sc_quit)
	{
		EventRecord	event;

		if (GI==NULL)
		{ /* no game */
			if (gamewindow!=NULL)
				KillVideo(); /* dispose of window and reset palette */
			if (WaitNextEvent(everyEvent,&event,15,NULL))
			{
				InitCursor();
				handleevent(&event,NULL);
			}

			if (getnextgamepath()!=NULL)
			{
				mycolorsinvalidate|= 2;
				loadgame(NULL);
				softclose = 0; /* it kept closing & reopening...; 2010.04.10 */
			}
			continue;
		}

		if (gamewindow==NULL)
		{ /* should have no window for no game */
			InitVideo(GI);
			InitSound(GI);
		}

		updatemenus(); //FCEU may have changed anything; 2010.03.23
		vmenubar = 20+((**gamegdevice).gdRect.top);

		do
		{
			static uint8	eventskip = 1, take;

			uint8		*gfx;
			int16		*sound;
			int32		ssize;
#ifdef env68k
			if (--eventskip==0)
#endif
			{ /* must take event regardless of return value; 2009.10.08 10:31 EDT */
				event.what = nullEvent;
				take+= 2*GetNextEvent(everyEvent,&event);
				handleevent(&event,NULL);
				eventskip = 2; /*EmulationSpeedRatio>>29;*/
			}

			if (suspended)
			{
				if (eoptions&EO_BGRUN)
					goto emulate;
				continue;
			}

			if ((updatemouse() || fullscreen) && gamewindow==myfrontwindow)
			{ /* hide cursor when within window (window covers screen in fs) */
				if (!mousehidden && event.where.v>=vmenubar)
					{HideCursor(); mousehidden = true;}
			} else if (mousehidden || take==0)
				{InitCursor(); mousehidden = false;}
		emulate:
			FCEUI_Emulate(&gfx,&sound,&ssize);

			if (softclose!=0)
			{
			close:
				closegame();
				if (softclose==_sc_reopengame)
					opengame(NULL);
				if (~softclose&(_sc_reopengame|_sc_closegame))
					softclose = 0;
				continue;
			}
			xbsave = gfx;
			FCEUD_Update(gfx,sound,ssize);
		} while(GI!=NULL);

		xbsave = NULL;
		stopsound();

		//everythin' showin' up lookin' all fresh
		updatemenus();
		ShowMenuBar();
		InitCursor(); mousehidden = false;
	}
exit: /* restoration is automatic (atexit) */
	cleanup();

	return result;
}

int main(int argc, char **argv)
{
#pragma unused(argc)
	initialize((long int)&argv);

	if(!FCEUI_Initialize())
	{ /* memory problem (no need for dialog) */
		/*showalert(alertid_error,"\pFCEUI_Initialize failed!","\p","\p","\p");*/
		return 1;
	}

	if (!reinitmenus())
	{
		SysBeep(20);
		FCEUD_Message("## Some menus are missing?\n");
	}

	soundavailable = adequatesoundmanager();

	/* allow sound to be generated anyway; 2009.01.25 */
	/* enablemenu(0302,3,soundon); // sound quality, volume, channels */

	if (handlersinstalled)
		tryevents();
#ifdef CONFIGFILE_PATH
	configfile_load(CONFIGFILE_PATH);
#endif
	FCEUI_SetVidSystem(palyo&=1);
	FCEUI_SetGameGenie(genie&=1);

	fullscreen&= 1;

	setdirectories();

	/* update everything that may have been changed */

	if (eoptions&EO_CPALETTE)
		FCEUI_SetPaletteArray(cpalette);
	else
		FCEUI_SetPaletteArray(NULL);

	FCEUI_SetSoundVolume(soundvolume);
	if (soundon && soundquality!=0)
		FCEUI_SetSoundQuality(soundquality);
	FCEUI_SetRectFilterAmp(rectangle_amp); //2010.03.23

	FCEUI_SetNTSCTH(ntsccol,ntsctint,ntschue);

	/* and lastly... */
	videorefresh();

	return eventloop();
}

void FCEUD_Update(uint8 *XBuf,
	int16 *sampledata,
	int samplecount)
{
	static int	skipcount = 0;
	int		maxskip = maxconbskip;
	uint8		skip = NoWaiting&1;

	if (skip)
		maxskip = ffbskip;

	if (!soundon || eoptions&EO_MUTE && (!(eoptions&EO_BGRUN) || suspended))
		samplecount = 0;
	else if (samplecount!=0)
		writesound(sampledata,samplecount); /* does the waiting itself */

	if (skip && skipcount<maxskip)
		{++skipcount; goto input;}
	skipcount = 0;

	if (samplecount==0)
		SpeedThrottle();

	if (XBuf!=NULL)
	{
		FCEUD_BlitScreen(XBuf);

		if (fceutileview)
			BlitTileWindow(0);
	}
input:
	if (!suspended)
		FCEUD_UpdateInput();
}
