/*
 macos/throttle.c
  for timing functions

  Created: 2008.08.12
  Revised: 2010.04.01

Saturday, March 27, 2010 5:30 PM
  Bits of NoWaiting now are 8, then 4 to avoid conflict with netplay.c.
*/

#include <Timer.h>

#include "shell.h"

/* ---- timer */

uint64 FCEUD_GetTime()
{
	uint64		time;

	Microseconds((UnsignedWide*)&time);
	return time;
}

uint64 FCEUD_GetTimeFreq()
{
	uint64		time;

	_int64xset(time,1000000);
	return time;
}

/* ---- throttle */

static int32		frametime;
static uint8		hframetime; /* for adding halves (slightly more accurate than before) */

void RefreshThrottleFPS()
{
	uint64		numer;
	uint32		denom;

	if ((denom=FCEUI_GetDesiredFPS())==0)
		denom = 1;
	_int64xset(numer,1000000*31250000.0); /* previously 1000000*16777216.0 */
	frametime = uint64divu32(numer,denom);
	hframetime = frametime&1, frametime>>= 1;
}

void SpeedThrottle()
{
	static uint64	ttime;
	static uint32	ltime = 0; /* allow wrap in case anything goes wrong */
	static uint8	htime = 0;
	int32		diff, calctime = frametime+(htime^=hframetime);

waiter:
	Microseconds((UnsignedWide*)&ttime);
	/* fixed the timer bug; 2009.09.19 23:14 */
	if ((diff=(signed)int64lo(ttime)-(signed)ltime>>2)<0)
	{
		diff = -diff;
		if (diff>calctime)
			goto set;
		goto waiter;
	}
	if (diff>calctime) /* four frames ahead?  reset */
	{
	set:
		ltime = int64lo(ttime);

		if (NoWaiting&8) /* consistent slowdown; overrun sound playback to catch up */
			NoWaiting|= 4; /* computer may be too slow for real-time */

		NoWaiting|= 8;
	} else /* otherwise advance frame time */
	{
		ltime+= calctime;
		NoWaiting&=~(4|8);
	}
}
