/*
 tools/MFAsm.c

C -o ::bin:MFAsm :tools:MFAsm.o :obj:cart.c.o :obj:disasm6502.c.o :obj:file.c.o 
	:obj:iNES.c.o :obj:interpreter.c.o
*/

#include <stdio.h>
#include <errno.h>

#include "../driver.h"
#include "../cart.h"
#include "disasm6502.h"
#include "../FCEU.h"
#include "../file.h"
#include "../iNES.h"
#include "interpreter.h"

#ifdef mac
 #define signalbreak()	((*((volatile UInt8*)0x179)&*((volatile UInt8*)0x17A))!=0)
#endif

static const char	*pname;

enum
{ /* err codes */
	_missing_param = -5,
	_unknown_param = -4,
	_bad_param = -3,
	_not_iNES = 2,
	_zero_ROM
};

static void PrintErr(const char *n,
	int t)
{
	char		b[32];
	const char	*s;

	switch(t)
	{
	case _missing_param: s = "parameter required"; break;
	case _unknown_param: s = "unknown option"; break;
	case _bad_param: s = "bad parameter"; break;
	case -2: s = "memory error"; break;
	case -1: s = "file or internal error"; break;
	case 0: s = "error opening file"; break;
	case _not_iNES: s = "this tool only DisAsm's iNES files"; break;
	case _zero_ROM: s = "ROM_size is zero"; break;

	default: s = "file error";
	}

	if (errno)
		sprintf(b,"; errno: %d",
			errno);
	else
		*b = '\0';

	if (n)
		fprintf(stderr,"# %s: %s, '%s'%s",
			pname,s,n,b);
	else
		fprintf(stderr,"# %s: %s%s",
			pname,s,b);
end:
	if (t<0)
		exit(1);
}

static uint32	absEnd;

static DECLFR(iNESROMRead)
{
extern uint8 *iNESPRGROM;
	if (A>=absEnd || iNESPRGROM==0)
		return A>>8;
	return iNESPRGROM[A];
}

static readfunc	ShortARead[1] = iNESROMRead;

static void DisassembleFile(const char *n,
	uint32 s,
	uint32 e)
{
	char		buf[32];
	FCEUFILE	*fp;
	int		r;

	if ((fp=FCEU_fopen(n,0,"rb",0))==0)
		PrintErr(n,~0);
	if ((r=iNESRead(n))<=0)
		{PrintErr(n,r==0 ? _not_iNES:r); return;}
	if (ROM_size==0)
		{PrintErr(n,_zero_ROM); goto end;}
	absEnd = ROM_size<<13;
	if (e>absEnd)
		e = absEnd;
	while(s<e)
		s = X6502_Disassemble(0,ShortARead,0,s,31,buf),
		printf("% 6X %s",
			s,buf);
end:
	iNESGI(GI_CLOSE);
}

static void ParseParameters(char const*const*argv,
	char const*const*e)
{
	char const*const*a = argv;

	long int	starto = 0, endo = -1U;

	while(a<e)
	{
		char const	*s = *a++;

		if (!strcmp(s,"-s"))
		{
			if (a==e)
				PrintErr(s,_missing_param);
			else if (readanyinteger(*a,strlen(*a),&starto)==0
					|| starto<0)
				PrintErr(s,_bad_param);
		} else if (!strcmp(s,"-e"))
		{
			if (a==e)
				PrintErr(s,_missing_param);
			else if (readanyinteger(*a,strlen(*a),&endo)==0)
				PrintErr(s,_bad_param);
		} else if (!strcmp(s,"-d"))
			if (a==e)
				PrintErr(s,_missing_param);
			else
				DisassembleFile(*a,starto,endo);
		else
			PrintErr(s,_unknown_param);
	}
}

void main(int argc,
	char const*const*argv)
{
	if (argc<=0)
		return;
	pname = *argv++;
	if (--argc==0)
	{
		fprintf(stderr,"# Usage: '%s' [-s startAddr] [-e endAddr] params...\n",
			pname);
		puts(
"# -d    disassemble file",
			stderr);
		return;
	}
	ParseParameters(argv,argv+argc);
}
